""" Implementation of some data types used in Participant Model"""
import decimal
from collections import namedtuple
import traceback
import inspect
import itertools
import enum
import warnings
from typing import Union
from lxml import etree as etree_
from sdc11073 import namespaces
from .mdib import containerproperties as cp
from math import isclose

"""
Interface of  pmTypes:

fromEtreeNode (class method) is a constructor that is used to create a type object from XML node

asEtreeNode: returns an etree node that represents the object
"""


class PropertyBasedPMType(object):
    """ Base class that assumes all data is defined as containerproperties and _props lists all property names."""

    def asEtreeNode(self, qname, nsmap):
        node = etree_.Element(qname, nsmap=nsmap)
        self._updateNode(node)
        return node

    def _updateNode(self, node):
        for prop_name, prop in self._sortedContainerProperties():
            try:
                prop.updateXMLValue(self, node)
            except Exception as ex:
                raise RuntimeError(
                    'In {}.{}, {} could not update: {}'.format(self.__class__.__name__, prop_name, str(prop),
                                                               traceback.format_exc()))

    def updateFromNode(self, node):
        for dummy, prop in self._sortedContainerProperties():
            prop.updateFromNode(self, node)

    def _sortedContainerProperties(self):
        """
        @return: a list of (name, object) tuples of all GenericProperties ( and subclasses)
        list is created based on _props lists of classes
        """
        ret = []
        classes = inspect.getmro(self.__class__)
        for cls in reversed(classes):
            try:
                names = cls._props  # pylint: disable=protected-access
            except:
                continue
            for name in names:
                obj = getattr(cls, name)
                if obj is not None:
                    ret.append((name, obj))
        return ret

    def __eq__(self, other):
        """ compares all properties"""
        try:
            for name, dummy in self._sortedContainerProperties():
                my_value = getattr(self, name)
                other_value = getattr(other, name)
                if my_value == other_value:
                    continue
                elif (isinstance(my_value, float) or isinstance(other_value, float)) and isclose(my_value, other_value):
                    continue  # float compare (almost equal)
                else:
                    return False
            return True
        except (TypeError, AttributeError):
            return False

    def __ne__(self, other):
        return not self == other

    @classmethod
    def fromNode(cls, node):
        """ default fromNode Constructor that provides no arguments for class __init__"""
        obj = cls()
        obj.updateFromNode(node)
        return obj


class ElementWithTextOnly(PropertyBasedPMType):
    text = cp.NodeTextProperty()  # this is the text of the node. Here attribute is lower case!
    _props = ['text']
    """An Element that has no attributes, only a text."""

    def __init__(self, text):
        self.text = text

    @classmethod
    def fromNode(cls, node):
        text = node.text
        return cls(text)


class LocalizedText(PropertyBasedPMType):
    text = cp.NodeTextProperty()  # this is the text of the node. Here attribute is lower case!
    Ref = cp.NodeAttributeProperty('Ref')
    Lang = cp.NodeAttributeProperty('Lang')
    Version = cp.IntegerAttributeProperty('Version')
    TextWidth = cp.NodeAttributeProperty('TextWidth')  # one of xs, s, m, l, xl, xxl
    _props = ['text', 'Ref', 'Lang', 'Version', 'TextWidth']
    """ Represents a LocalizedText type in the Participant Model. """

    def __init__(self, text, lang=None, ref=None, version=None, textWidth=None):
        """
        :param text: a string
        :param lang: a string or None
        :param ref: a string or None
        :param version: an int or None
        :param textWidth: xs, s, m, l, xl, xxl or None
        """
        self.text = text
        self.Lang = lang
        self.Ref = ref
        self.Version = version
        self.TextWidth = textWidth

    def __repr__(self):
        repr_string = 'LocalizedText("{}"'.format(self.text)
        if self.Lang is not None:
            repr_string += ', lang="{}"'.format(self.Lang)
        if self.Ref is not None:
            repr_string += ', ref="{}"'.format(self.Ref)
        if self.Version is not None:
            repr_string += ', version="{}"'.format(self.Version)
        if self.TextWidth is not None:
            repr_string += ', textWidth="{}"'.format(self.TextWidth)
        repr_string += ')'
        return repr_string

    @classmethod
    def fromNode(cls, node):
        text = node.text
        lang = node.get('Lang')
        ref = node.get('Ref')
        version = node.get('Version')
        if version is not None:
            version = int(version)
        textWidth = node.get('TextWidth')
        return cls(text, lang, ref, version, textWidth)


DefaultCodingSystem = 'urn:oid:1.2.840.10004.1.1.1.0.0.1'  # ISO/IEC 11073-10101


def _get_default_coding_system():
    """Helper method that allows handling of DefaultCodingSystem modification at runtime.
    By using it implied coding system always is the current value of DefaultCodingSystem."""
    return DefaultCodingSystem


class NotCompareableVersionError(Exception):
    """ This exception says that two coded values cannot be compared, because one has a coding system version, the other one not.
    In that case it is not possible to decide if they are equal."""
    pass


_CodingBase = namedtuple('_CodingBase', 'code codingSystem codingSystemVersion')


class Coding(_CodingBase):
    """ Immutable representation of a coding. Can be used as key in dictionaries"""

    def __new__(cls, code, codingSystem=None, codingSystemVersion=None):
        return super(Coding, cls).__new__(cls,
                                          str(code),
                                          codingSystem or DefaultCodingSystem,
                                          codingSystemVersion)

    def equals(self, other, raiseNotComparableException=False):
        """ different compare method to __eq__, overwriting this one makes Coding unhashable!
         other can be an int, a string, or a Coding.
         Simple comparison with int or string only works if self.codingSystem == DefaultCodingSystem
         and self.codingSystemVersion is None"""
        if isinstance(other, int):
            other = str(other)
        if isinstance(other, str):
            # compare to 11073 coding system where codes are strings
            if self.codingSystem != DefaultCodingSystem:
                return False
            elif self.code != other:
                return False
            elif self.codingSystemVersion is None:
                return True
            else:
                if raiseNotComparableException:
                    raise NotCompareableVersionError(
                        'no simple compare, self.codingSystemVersion == {}'.format(self.codingSystemVersion))
                else:
                    return False
        else:
            try:
                if self.code != other.code or self.codingSystem != other.codingSystem:
                    return False
                elif self.codingSystemVersion == other.codingSystemVersion:
                    return True
                elif self.codingSystemVersion is None or other.codingSystemVersion is None:
                    if raiseNotComparableException:
                        raise NotCompareableVersionError(
                            'my codingSystem = "{}", other = "{}"'.format(self.codingSystemVersion,
                                                                          other.codingSystemVersion))
                    else:
                        return False
                else:
                    return False
            except AttributeError:
                return False

    @classmethod
    def fromNode(cls, node):
        """ Read Code and CodingSystem attributes of a node (CodedValue). """
        code = node.get('Code')
        codingSystem = node.get('CodingSystem', DefaultCodingSystem)
        codingSystemVersion = node.get('CodingSystemVersion')
        return cls(code, codingSystem, codingSystemVersion)


def mkCoding(code, codingSystem=None, codingSystemVersion=None):
    return Coding(code, codingSystem or DefaultCodingSystem, codingSystemVersion)


class T_Translation(PropertyBasedPMType):
    """
    Translation is part of CodedValue in BICEPS FINAL
    """
    ext_Extension = cp.ExtensionNodeProperty()
    Code = cp.NodeAttributeProperty('Code')
    CodingSystem = cp.NodeAttributeProperty('CodingSystem', impliedPyValue=_get_default_coding_system)
    CodingSystemVersion = cp.NodeAttributeProperty('CodingSystemVersion')

    _props = ['ext_Extension', 'Code', 'CodingSystem', 'CodingSystemVersion']

    def __init__(self, code=None, codingsystem=None, codingSystemVersion=None):
        """

        :param code: a string or an int
        :param codingsystem: anyURI or None, defaults to ISO/IEC 11073-10101 if None
        :param codingSystemVersion: a string, min. length = 1
        """
        self.Code = str(code)
        self.CodingSystem = codingsystem
        self.CodingSystemVersion = codingSystemVersion
        self.coding = None  # oberwritten by self._mkCoding()
        self.mkCoding()

    def __repr__(self):
        if self.CodingSystem is None:
            return 'T_Translation("{}")'.format(self.Code)
        elif self.CodingSystemVersion is None:
            return 'T_Translation("{}", codingsystem="{}")'.format(self.Code, self.CodingSystem)
        else:
            return 'T_Translation("{}", codingsystem="{}", codingsystemversion="{}")' \
                .format(self.Code, self.CodingSystem, self.CodingSystemVersion)

    def mkCoding(self):
        if self.Code is not None:
            self.coding = Coding(self.Code, self.CodingSystem, self.CodingSystemVersion)
        else:
            self.coding = None

    def __eq__(self, other):
        """ other can be an int, a string, a CodedValue like object (has "coding" member) or a Coding"""
        if hasattr(other, 'coding'):
            return self.coding.equals(other.coding)
        else:
            return self.coding.equals(other)

    @classmethod
    def fromNode(cls, node):
        obj = cls(None)
        obj.updateFromNode(node)
        obj.mkCoding()
        return obj


class _CodedValueBase(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    CodingSystemName = cp.SubElementListProperty([namespaces.domTag('CodingSystemName')], cls=LocalizedText)
    ConceptDescription = cp.SubElementListProperty([namespaces.domTag('ConceptDescription')], cls=LocalizedText)
    Code = cp.NodeAttributeProperty('Code')
    CodingSystem = cp.NodeAttributeProperty('CodingSystem', impliedPyValue=_get_default_coding_system)
    CodingSystemVersion = cp.NodeAttributeProperty('CodingSystemVersion')
    SymbolicCodeName = cp.NodeAttributeProperty('SymbolicCodeName')
    Type = cp.XsiTypeAttributeProperty(namespaces.QN_TYPE)
    _props = ['ext_Extension', 'CodingSystemName', 'ConceptDescription',
              'Code', 'CodingSystem', 'CodingSystemVersion', 'SymbolicCodeName', 'Type']
    # aliases for backward compatibility
    codingSystemNames = CodingSystemName
    conceptDescriptions = ConceptDescription
    codingSystem = CodingSystem

    def __init__(self, code, codingsystem=None, codingSystemVersion=None, codingSystemNames=None,
                 conceptDescriptions=None, symbolicCodeName=None):
        """
        :param code: a string or an int
        :param codingsystem: anyURI or None, defaults to ISO/IEC 11073-10101 if None
        :param codingSystemVersion: a string, min. length = 1
        :param codingSystemNames: a list of LocalizedText objects or None
        :param conceptDescriptions: a list of LocalizedText objects or None
        :param symbolicCodeName: a string, min. length = 1 or None
        """
        self.Code = str(code)
        self.CodingSystem = codingsystem
        self.CodingSystemVersion = codingSystemVersion
        self.CodingSystemName = [] if codingSystemNames is None else codingSystemNames
        self.ConceptDescription = [] if conceptDescriptions is None else conceptDescriptions
        self.SymbolicCodeName = symbolicCodeName
        self.coding = None  # oberwritten by self._mkCoding()
        self.mkCoding()

    def mkCoding(self):
        if self.Code is not None:
            self.coding = Coding(self.Code, self.CodingSystem, self.CodingSystemVersion)
        else:
            self.coding = None

    def __repr__(self):
        if self.CodingSystem is None:
            return 'CodedValue("{}")'.format(self.Code)
        elif self.CodingSystemVersion is None:
            return 'CodedValue("{}", codingsystem="{}")'.format(self.Code, self.CodingSystem)
        else:
            return 'CodedValue("{}", codingsystem="{}", codingsystemversion="{}")' \
                .format(self.Code, self.CodingSystem, self.CodingSystemVersion)

    @classmethod
    def fromNode(cls, node):
        obj = cls(None)
        obj.updateFromNode(node)
        obj.mkCoding()
        return obj


class CodedWithTranslations(_CodedValueBase):
    Translation = cp.SubElementListProperty([namespaces.domTag('Translation')], cls=_CodedValueBase)
    _props = ['Translation']

    def __eq__(self, other):
        """ other can be an int, a string, a CodedValue like object (has "coding" member) or a Coding"""
        if hasattr(other, 'coding'):
            return self.coding == other.coding
        else:
            return self.coding == other


class CodedValue(_CodedValueBase):
    Translation = cp.SubElementListProperty([namespaces.domTag('Translation')], cls=T_Translation)
    _props = ['Translation']

    def __eq__(self, other):
        """ This operator handles not comparable versions as different versions

        :param other: int, str or another CodedValue
        """
        if isinstance(other, (int, str)):
            return have_matching_codes(self, Coding(other))
        return have_matching_codes(self, other)

    def equals(self, other, raiseNotComparableException=True):
        """
        Compare this CodedValue with another one.
        A simplified compare with an int or string is also possible, it assumes DefaultCodingSystem and no CodingSystemVersion
        :param other: int, str or another CodedValue
        :param raiseNotComparableException: if False, not comparable versions are handled as different versions
                        if True, a NotCompareableVersionError is thrown if any of the Codings are not comparable
        :return: boolean
        """
        warnings.warn('equals is deprecated, use function have_matching_codes', DeprecationWarning)
        if isinstance(other, self.__class__):
            # Two CodedValue objects C1 and C2 are equivalent, if there exists a CodedValue object T1 in C1/pm:Translation
            # and a CodedValue object T2 in C2/pm:Translation such that T1 and T2 are equivalent, C1 and T2 are equivalent, or C2 and T1 are equivalent.
            foundMatch = False
            not_comparables = []
            my_codes = [self]  # C1
            my_codes.extend(self.Translation)  # all T1
            other_codes = [other]  # C2
            other_codes.extend(other.Translation)  # all T2
            for left, right in itertools.product(my_codes, other_codes):
                try:
                    if left.coding.equals(right.coding, raiseNotComparableException):
                        if not raiseNotComparableException:
                            return True
                        else:
                            foundMatch = True
                except NotCompareableVersionError as ex:
                    not_comparables.append(str(ex))
            if not_comparables:
                raise NotCompareableVersionError(';'.join(not_comparables))
            else:
                return foundMatch
        else:
            # simplified compare: compare to 11073 coding system where codes are integers, ignore translations
            return self.coding.equals(other, raiseNotComparableException)

    @classmethod
    def fromNode(cls, node):
        nodeType = None
        nodeTypeText = node.get(namespaces.QN_TYPE)
        if nodeTypeText is not None:
            nodeType = namespaces.txt2QName(node.get(namespaces.QN_TYPE), node.nsmap)
        if nodeType == namespaces.domTag('CodedWithTranslations'):
            cls = CodedWithTranslations  # pylint: disable=self-cls-assignment

        obj = cls(None)
        obj.updateFromNode(node)
        obj.mkCoding()
        return obj


def have_matching_codes(code_a: Union[CodedValue, Coding], code_b: Union[CodedValue, Coding]) -> bool:
    """A CodedValue is a set of codings (coding plus translations), a Coding is a set with only one element.
     Function returns false if no coding is found in both sets, otherwise True."""
    codes_a = set()
    codes_b = set()
    for the_set, the_code in [(codes_a, code_a), (codes_b, code_b)]:
        try:
            the_set.add(the_code.coding)
            if the_code.Translation is not None:
                for tr in the_code.Translation:
                    the_set.add(tr.coding)
        except AttributeError:
            the_set.add(the_code)
    common_codes = codes_a.intersection(codes_b)
    return len(common_codes) > 0


class Annotation(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    Type = cp.SubElementProperty([namespaces.domTag('Type')], valueClass=CodedValue)
    _props = ['ext_Extension', 'Type']

    codedValue = Type
    """ An Annotation contains a Type Element that is a CodedValue.
    This is intended as an immutable object. After it has been created, no modification shall be done. """

    def __init__(self, codedValue):
        self.Type = codedValue
        self.coding = codedValue.coding

    @classmethod
    def fromNode(cls, node):
        typeNode = node.find(namespaces.domTag('Type'))
        codedValue = CodedValue.fromNode(typeNode)
        return cls(codedValue)


class OperationGroup(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    Type = cp.SubElementProperty([namespaces.domTag('Type')], valueClass=CodedValue)
    OperatingMode = cp.NodeAttributeProperty('OperatingMode')
    Operations = cp.NodeAttributeListProperty('Operations')  # pm:OperationRef
    _props = ['ext_Extension', 'Type', 'OperatingMode', 'Operations']

    def __init__(self, codedValue=None, operatingMode=None, operations=None):
        """
        :param codedValue: a CodedValue instances or None
        :param operatingMode:  xsd:string string
        :param operations: a xsd:string
        """
        self.Type = codedValue
        self.OperatingMode = operatingMode
        self.Operations = operations

    @classmethod
    def fromNode(cls, node):
        ret = cls()
        ret.updateFromNode(node)
        return ret


class InstanceIdentifier(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    Type = cp.SubElementProperty([namespaces.domTag('Type')], valueClass=CodedValue)
    IdentifierName = cp.SubElementListProperty([namespaces.domTag('IdentifierName')], cls=LocalizedText)
    Root = cp.NodeAttributeProperty('Root',
                                    defaultPyValue='biceps.uri.unk')  # xsd:anyURI string, default is defined in R0135
    Extension = cp.NodeAttributeProperty('Extension')  # a xsd:string
    _props = ('ext_Extension', 'Type', 'IdentifierName', 'Root', 'Extension')

    def __init__(self, root, type_codedValue=None, identifierNames=None, extensionString=None):
        """
        :param root:  xsd:anyURI string
        :param type_codedValue: a CodedValue instances or None
        :param identifierNames: a list of LocalizedText instances or None
        :param extensionString: a xsd:string
        """
        self.Root = root
        self.Type = type_codedValue
        self.IdentifierName = [] if identifierNames is None else identifierNames
        self.Extension = extensionString
        self.node = None

    @classmethod
    def fromNode(cls, node):
        root = None
        extensionString = None
        type_codedValue = None
        identifierNames = None
        ret = cls(root, type_codedValue=type_codedValue, identifierNames=identifierNames,
                  extensionString=extensionString)
        ret.updateFromNode(node)
        ret.node = node
        return ret

    def __repr__(self):
        return 'InstanceIdentifier(root={!r}, Type={} ext={!r})'.format(self.Root, self.Type, self.Extension)


class Range(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    Lower = cp.DecimalAttributeProperty('Lower')  # optional, an integer or float
    Upper = cp.DecimalAttributeProperty('Upper')  # optional, an integer or float
    StepWidth = cp.DecimalAttributeProperty('StepWidth')  # optional, an integer or float
    RelativeAccuracy = cp.DecimalAttributeProperty('RelativeAccuracy')  # optional, an integer or float
    AbsoluteAccuracy = cp.DecimalAttributeProperty('AbsoluteAccuracy')  # optional, an integer or float
    _props = ['ext_Extension', 'Lower', 'Upper', 'StepWidth', 'RelativeAccuracy', 'AbsoluteAccuracy']

    def __init__(self, lower=None, upper=None, stepWidth=None, relativeAccuracy=None, absoluteAccuracy=None):
        """
        :param lower: The including lower bound of the range. A value as float or integer, can be None
        :param upper: The including upper bound of the range. A value as float or integer, can be None
        :param stepWidth: The numerical distance between two values in the range of the given upper and lower bound. A value as float or integer, can be None
        :param relativeAccuracy: Maximum relative error in relation to the correct value within the given range. A value as float or integer, can be None
        :param absoluteAccuracy: Maximum absolute error in relation to the correct value within the given range. A value as float or integer, can be None
        """
        self.Lower = lower
        self.Upper = upper
        self.StepWidth = stepWidth
        self.RelativeAccuracy = relativeAccuracy
        self.AbsoluteAccuracy = absoluteAccuracy

    def __repr__(self):
        return 'Range (Lower={!r}, Upper={!r}, StepWidth={!r}, RelativeAccuracy={!r}, AbsoluteAccuracy={!r})' \
            .format(self.Lower, self.Upper, self.StepWidth, self.RelativeAccuracy, self.AbsoluteAccuracy)


class Measurement(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    MeasurementUnit = cp.SubElementProperty([namespaces.domTag('MeasurementUnit')], valueClass=CodedValue)  # mandatory
    MeasuredValue = cp.DecimalAttributeProperty('MeasuredValue')
    _props = ['ext_Extension', 'MeasurementUnit', 'MeasuredValue']

    def __init__(self, value, unit):
        """
        :param value: a value as string, float or integer
        :param unit: a CodedValue instance
        """
        self.MeasuredValue = value
        self.MeasurementUnit = unit

    @classmethod
    def fromNode(cls, node):
        value = node.get('MeasuredValue')
        if value is not None:
            value = float(value)
        unit = None
        unitNode = node.find(namespaces.domTag('MeasurementUnit'))
        if unitNode is not None:
            unit = CodedValue.fromNode(unitNode)
        return cls(value, unit)

    def __repr__(self):
        return 'Measurement(value={!r}, Unit={!r})'.format(self.MeasuredValue, self.MeasurementUnit)


class AllowedValue(PropertyBasedPMType):
    Value = cp.NodeTextProperty([namespaces.domTag('Value')], isOptional=False)
    Type = cp.SubElementProperty([namespaces.domTag('Type')], valueClass=CodedValue)
    _props = ['Value', 'Type']
    typeCoding = Type
    value = Value

    def __init__(self, valueString, typeCoding=None):
        """One AllowedValue of a EnumStringMetricDescriptor. It has up to two sub elements "Value" and "Type"(optional).
        A StringEnumMetricDescriptor has a list of AllowedValues.
        :param valueString: a string
        :param typeCoding: an optional CodedValue instance
        """
        self.Value = valueString
        self.Type = typeCoding

    @classmethod
    def fromNode(cls, node):
        valueString = node.find(namespaces.domTag('Value')).text
        typeNode = node.find(namespaces.domTag('Type'))
        if typeNode is None:
            typeCoding = None
        else:
            typeCoding = CodedValue.fromNode(typeNode)
        return cls(valueString, typeCoding)


class AbstractMetricValue(PropertyBasedPMType):
    """ This is the base class for metric values inside metric states"""
    ext_Extension = cp.ExtensionNodeProperty()
    StartTime = cp.TimestampAttributeProperty('StartTime')  # time.time() value (float)
    StopTime = cp.TimestampAttributeProperty('StopTime')  # time.time() value (float)
    DeterminationTime = cp.TimestampAttributeProperty('DeterminationTime')  # time.time() value (float)
    MQ_Extension = cp.ExtensionNodeProperty([namespaces.domTag('MetricQuality')])
    Validity = cp.NodeAttributeProperty('Validity', [namespaces.domTag('MetricQuality')])  # pm:MeasurementValidity
    Mode = cp.NodeAttributeProperty('Mode', [namespaces.domTag('MetricQuality')],
                                    impliedPyValue='Real')  # pm:GenerationMode
    Qi = cp.DecimalAttributeProperty('Qi', [namespaces.domTag('MetricQuality')],
                                     impliedPyValue=1)  # pm:QualityIndicator
    Annotation = cp.SubElementListProperty([namespaces.domTag('Annotation')], Annotation)
    _props = ('ext_Extension', 'StartTime', 'StopTime', 'DeterminationTime', 'MQ_Extension', 'Validity', 'Mode', 'Qi',
              'Annotation')

    Annotations = Annotation

    def __init__(self, nsmapper, node=None):
        self._nsmapper = nsmapper
        # attributes of root node
        self.node = node
        if node is not None:
            self.updateFromNode(node)
        else:
            self.Validity = 'Vld'  # mandatory value, for convenience it is preset to Vld

    def updateFromNode(self, node):
        for dummy, prop in self._sortedContainerProperties():
            prop.updateFromNode(self, node)
        self.node = node

    def asEtreeNode(self, qname, nsmap):
        node = super(AbstractMetricValue, self).asEtreeNode(qname, nsmap)
        node.set(namespaces.QN_TYPE, namespaces.docNameFromQName(self.QType, nsmap))
        return node

    @classmethod
    def fromNode(cls, node):
        obj = cls(node.nsmap, node)
        return obj


class NumericMetricValue(AbstractMetricValue):
    QType = namespaces.domTag('NumericMetricValue')
    Value = cp.DecimalAttributeProperty('Value')  # an integer or float
    _props = ('Value',)

    def __repr__(self):
        return '{} Validity={} Value={} DeterminationTime={}'.format(self.__class__.__name__,
                                                                     self.Validity,
                                                                     self.Value,
                                                                     self.DeterminationTime)

    def __eq__(self, other):
        """ compares all properties, special handling of Value member"""
        try:
            for name, dummy in self._sortedContainerProperties():
                if name == 'Value':
                    my_value = getattr(self, name)
                    other_value = getattr(other, name)
                    if my_value is None and other_value is None:
                        continue
                    elif my_value is None or other_value is None:
                        return False  # only one is None, they are not equal
                    else:
                        if decimal.Decimal(str(my_value)) != decimal.Decimal(str(other_value)):
                            return False
                else:
                    if getattr(self, name) == getattr(other, name):
                        continue
                    else:
                        return False
            return True
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other


class StringMetricValue(AbstractMetricValue):
    QType = namespaces.domTag('StringMetricValue')
    Value = cp.NodeAttributeProperty('Value')  # a string
    _props = ('Value',)

    def __repr__(self):
        return '{} Validity={} Value={} DeterminationTime={}'.format(self.__class__.__name__,
                                                                     self.Validity,
                                                                     self.Value,
                                                                     self.DeterminationTime)


class ApplyAnnotation(PropertyBasedPMType):
    AnnotationIndex = cp.IntegerAttributeProperty('AnnotationIndex')
    SampleIndex = cp.IntegerAttributeProperty('SampleIndex')
    _props = ['AnnotationIndex', 'SampleIndex']

    def __init__(self, annotationIndex, sampleIndex):
        self.AnnotationIndex = annotationIndex
        self.SampleIndex = sampleIndex

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        cls.AnnotationIndex.updateFromNode(obj, node)
        cls.SampleIndex.updateFromNode(obj, node)
        return obj

    def __repr__(self):
        return '{} AnnotationIndex={} SampleIndex={}'.format(self.__class__.__name__, self.AnnotationIndex,
                                                             self.SampleIndex)


class SampleArrayValue(AbstractMetricValue):
    QType = namespaces.domTag('SampleArrayValue')
    Samples = cp.DecimalListAttributeProperty('Samples')  # list of xs:decimal types
    ApplyAnnotations = cp.SubElementListProperty([namespaces.domTag('ApplyAnnotation')], ApplyAnnotation)
    _props = ('Samples', 'ApplyAnnotations')

    def __repr__(self):
        return '{} Samples={} ApplyAnnotations={}'.format(self.__class__.__name__, self.Samples, self.ApplyAnnotations)

    def __eq__(self, other):
        """ compares all properties, special handling of Value member"""
        try:
            for name, dummy in self._sortedContainerProperties():
                if name == 'Samples':
                    own_sample = getattr(self, name)
                    other_sample = getattr(other, name)
                    # avoid None values; treat None and empty list as equal
                    if own_sample is None:
                        own_sample = []
                    if other_sample is None:
                        other_sample = []
                    if len(own_sample) != len(other_sample):
                        return False
                    for pair in zip(own_sample, other_sample):
                        if decimal.Decimal(str(pair[0])) != decimal.Decimal(str(pair[1])):
                            return False
                else:
                    if getattr(self, name) == getattr(other, name):
                        continue
                    else:
                        return False
            return True
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other


class RemedyInfo(PropertyBasedPMType):
    """An Element that has
         0..1 Subelement "Extension" (not handled here)
         0..n SubElements "Description" type=pm:LocalizedText."""
    ext_Extension = cp.ExtensionNodeProperty()
    Description = cp.SubElementListProperty([namespaces.domTag('Description')], cls=LocalizedText)
    _props = ['ext_Extension', 'Description']

    def __init__(self, descriptions=None):
        """
        :param descriptions : a list of LocalizedText objects or None
        """
        if descriptions:
            self.Description = descriptions


class CauseInfo(PropertyBasedPMType):
    """An Element that has
         0..1 Subelement "RemedyInfo", type = pm:RemedyInfo
         0..n SubElements "Description" type=pm:LocalizedText."""
    ext_Extension = cp.ExtensionNodeProperty()
    RemedyInfo = cp.SubElementProperty([namespaces.domTag('RemedyInfo')], valueClass=RemedyInfo)
    Description = cp.SubElementListProperty([namespaces.domTag('Description')], cls=LocalizedText)
    _props = ['ext_Extension', 'RemedyInfo', 'Description']

    def __init__(self, remedyInfo, descriptions):
        """
        :param remedyInfo: a RemedyInfo instance or None
        :param descriptions : a list of LocalizedText objects or None
        """
        self.RemedyInfo = remedyInfo
        self.Description = descriptions

    @classmethod
    def fromNode(cls, node):
        remedyInfoNode = node.find(namespaces.domTag('RemedyInfo'))
        if remedyInfoNode is not None:
            remedyInfo = RemedyInfo.fromNode(remedyInfoNode)
        else:
            remedyInfo = None
        descriptions = []
        descriptionNodes = node.findall(namespaces.domTag('Description'))
        for d in descriptionNodes:
            descriptions.append(LocalizedText.fromNode(d))
        return cls(remedyInfo, descriptions)


class Argument(PropertyBasedPMType):
    """An Element that has
         1 Subelement "ArgName", type = pm:CodedValue
         1 SubElement "Arg" type=QName."""
    ArgName = cp.SubElementProperty([namespaces.domTag('ArgName')], valueClass=CodedValue)
    Arg = cp.NodeTextQNameProperty([namespaces.domTag('Arg')])
    _props = ['ArgName', 'Arg']

    def __init__(self, argName, arg):
        """
        :param argName: a CodedValue instance
        :param arg : etree_.QName instance
        """
        self.ArgName = argName
        self.Arg = arg

    @classmethod
    def fromNode(cls, node):
        argNameNode = node.find(namespaces.domTag('ArgName'))
        argName = CodedValue.fromNode(argNameNode)
        argNode = node.find(namespaces.domTag('Arg'))
        arg_QName = namespaces.txt2QName(argNode.text, node.nsmap)
        return cls(argName, arg_QName)

    def __repr__(self):
        return 'Argument(argName={}, arg={})'.format(self.ArgName, self.Arg)


class PhysicalConnectorInfo(PropertyBasedPMType):
    """PhysicalConnectorInfo defines a number in order to allow to guide the clinical user for a failure,
    e.g., in case of a disconnection of a sensor or an ultrasonic handpiece.
    Only in BICEPS final!"""
    ext_Extension = cp.ExtensionNodeProperty()
    Label = cp.SubElementListProperty([namespaces.domTag('Label')],
                                      cls=LocalizedText)  # A human-readable label that describes the physical connector.
    Number = cp.IntegerAttributeProperty('Number')  # Number designates the connector number of the physical connector.
    _props = ['ext_Extension', 'Label', 'Number']

    def __init__(self, labels, number):
        """
        :param labels: a  list of LocalizedText
        :param number : an integer
        """
        self.Label = labels
        self.Number = number

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        cls.Label.updateFromNode(obj, node)
        cls.Number.updateFromNode(obj, node)
        return obj

    def __repr__(self):
        return 'PhysicalConnectorInfo(label={}, number={})'.format(self.Label, self.Number)


class SystemSignalActivation(PropertyBasedPMType):
    Manifestation = cp.NodeAttributeProperty('Manifestation',
                                             defaultPyValue='Oth')  # required, pmtypes.AlertSignalManifestation
    State = cp.NodeAttributeProperty('State', defaultPyValue='On')  # required,  pmtypes.AlertActivation
    _props = ['Manifestation', 'State']

    def __init__(self, manifestation, state):
        """
        :param manifestation: a pmtypes.AlertSignalManifestation value
        :param state : a pmtypes.AlertActivation value
        """
        self.Manifestation = manifestation
        self.State = state

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        obj.updateFromNode(node)
        return obj

    def __repr__(self):
        return 'SystemSignalActivation(Manifestation={}, State={})'.format(self.Manifestation, self.State)


class ProductionSpecification(PropertyBasedPMType):
    SpecType = cp.SubElementProperty([namespaces.domTag('SpecType')], valueClass=CodedValue)
    ProductionSpec = cp.NodeTextProperty([namespaces.domTag('ProductionSpec')], isOptional=False)
    ComponentId = cp.SubElementProperty([namespaces.domTag('ComponentId')], valueClass=InstanceIdentifier)  # optional
    _props = ['SpecType', 'ProductionSpec', 'ComponentId']

    def __init__(self, spectype, productionspec, componentid=None):
        """
        :param spectype: a pmtypes.CodedValue value
        :param productionspec: a string
        :param componentid : a pmtypes.InstanceIdentifier value
        """
        self.SpecType = spectype
        self.ProductionSpec = productionspec
        self.ComponentId = componentid

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        obj.updateFromNode(node)
        return obj


class BaseDemographics(PropertyBasedPMType):
    Givenname = cp.NodeTextProperty([namespaces.domTag('Givenname')])
    Middlename = cp.SubElementListProperty([namespaces.domTag('Middlename')], cls=ElementWithTextOnly)  # 0...n Elements
    Familyname = cp.NodeTextProperty([namespaces.domTag('Familyname')])
    Birthname = cp.NodeTextProperty([namespaces.domTag('Birthname')])
    Title = cp.NodeTextProperty([namespaces.domTag('Title')])
    _props = ['Givenname', 'Middlename', 'Familyname', 'Birthname', 'Title']

    def __init__(self, givenname=None, middlenames=None, familyname=None, birthname=None, title=None):
        self.Givenname = givenname
        if isinstance(middlenames, str):
            self.Middlename = [middlenames]
        else:
            self.Middlename = middlenames or []
        self.Familyname = familyname
        self.Birthname = birthname
        self.Title = title


class PersonReference(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    Identification = cp.SubElementListProperty([namespaces.domTag('Identification')], cls=InstanceIdentifier)  # 1...n
    Name = cp.SubElementProperty([namespaces.domTag('Name')], valueClass=BaseDemographics)  # optional
    _props = ['ext_Extension', 'Identification', 'Name']

    def __init__(self, identifications=None, name=None):
        """
        :param identifications: a list of InstanceIdentifier objects
        :param name: a BaseDemographics object
        """
        if identifications:
            self.Identification = identifications
        self.Name = name


class LocationDetail(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    PoC = cp.NodeAttributeProperty('PoC')
    Room = cp.NodeAttributeProperty('Room')
    Bed = cp.NodeAttributeProperty('Bed')
    Facility = cp.NodeAttributeProperty('Facility')
    Building = cp.NodeAttributeProperty('Building')
    Floor = cp.NodeAttributeProperty('Floor')
    _props = ('ext_Extension', 'PoC', 'Room', 'Bed', 'Facility', 'Building', 'Floor')

    def __init__(self, poc=None, room=None, bed=None, facility=None, building=None, floor=None):
        self.PoC = poc
        self.Room = room
        self.Bed = bed
        self.Facility = facility
        self.Building = building
        self.Floor = floor


class LocationReference(PropertyBasedPMType):  # BICEPS Final
    Identification = cp.SubElementListProperty([namespaces.domTag('Identification')], cls=InstanceIdentifier)  # 1...n
    LocationDetail = cp.SubElementProperty([namespaces.domTag('LocationDetail')], valueClass=LocationDetail)  # optional
    _props = ['Identification', 'LocationDetail']

    def __init__(self, identifications=None, locationdetail=None):
        if identifications:
            self.Identification = identifications
        self.LocationDetail = locationdetail


class PersonParticipation(PersonReference):
    Role = cp.SubElementListProperty([namespaces.domTag('Role')], cls=CodedValue)  # 0...n
    _props = ['Role', ]

    def __init__(self, identifications=None, name=None, roles=None):
        super(PersonParticipation, self).__init__(identifications, name)
        if roles:
            self.Role = roles


class ClinicalInfo(PropertyBasedPMType):
    Type = cp.SubElementProperty([namespaces.domTag('Type')], valueClass=CodedValue)  # optional
    Description = cp.SubElementListProperty([namespaces.domTag('Description')], cls=LocalizedText)  # 0...n
    RelatedMeasurement = cp.SubElementListProperty([namespaces.domTag('RelatedMeasurement')], cls=Measurement)  # 0...n
    _props = ['Type', 'Description', 'RelatedMeasurement']

    def __init__(self, typecode=None, descriptions=None, relatedmeasurements=None):
        """
        :param typecode: a CodedValue Instance
        :param descriptions: a list of LocalizedText objects
        :param relatedmeasurements: a list of Measurement objects
        """
        self.Type = typecode
        if descriptions:
            self.Description = descriptions
        if relatedmeasurements:
            self.RelatedMeasurement = relatedmeasurements


class ImagingProcedure(PropertyBasedPMType):
    AccessionIdentifier = cp.SubElementProperty([namespaces.domTag('AccessionIdentifier')],
                                                valueClass=InstanceIdentifier)  # mandatory
    RequestedProcedureId = cp.SubElementProperty([namespaces.domTag('RequestedProcedureId')],
                                                 valueClass=InstanceIdentifier)  # mandatory
    StudyInstanceUid = cp.SubElementProperty([namespaces.domTag('StudyInstanceUid')],
                                             valueClass=InstanceIdentifier)  # mandatory
    ScheduledProcedureStepId = cp.SubElementProperty([namespaces.domTag('ScheduledProcedureStepId')],
                                                     valueClass=InstanceIdentifier)  # mandatory
    Modality = cp.SubElementProperty([namespaces.domTag('Modality')], valueClass=CodedValue)  # optional
    ProtocolCode = cp.SubElementProperty([namespaces.domTag('ProtocolCode')], valueClass=CodedValue)  # optional
    _props = ['AccessionIdentifier', 'RequestedProcedureId', 'StudyInstanceUid', 'ScheduledProcedureStepId',
              'Modality', 'ProtocolCode']

    def __init__(self, accessionidentifier, requestedprocedureid, studyinstanceuid, scheduledprocedurestepid,
                 modality=None, protocolcode=None):
        self.AccessionIdentifier = accessionidentifier
        self.RequestedProcedureId = requestedprocedureid
        self.StudyInstanceUid = studyinstanceuid
        self.ScheduledProcedureStepId = scheduledprocedurestepid
        self.Modality = modality
        self.ProtocolCode = protocolcode

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None, None, None)
        obj.updateFromNode(node)
        return obj


class OrderDetail(PropertyBasedPMType):
    Start = cp.NodeTextProperty([namespaces.domTag('Start')])  # optional, xsd:dateTime
    End = cp.NodeTextProperty([namespaces.domTag('End')])  # optional, xsd:dateTime
    Performer = cp.SubElementListProperty([namespaces.domTag('Performer')], cls=PersonParticipation)  # 0...n)
    Service = cp.SubElementListProperty([namespaces.domTag('Service')], cls=CodedValue)  # 0...n)
    ImagingProcedure = cp.SubElementListProperty([namespaces.domTag('ImagingProcedure')], cls=ImagingProcedure)
    _props = ['Start', 'End', 'Performer', 'Service', 'ImagingProcedure']

    def __init__(self, start=None, end=None, performer=None, service=None, imagingprocedure=None):
        """
        :param start: a xsd:DateTime string
        :param end: a xsd:DateTime string
        :param performer: a list of PersonParticipation objects
        :param service: a list of CodedValue objects
        :param imagingprocedure: a list of ImagingProcedure objects
        """
        self.Start = start
        self.End = end
        if performer:
            self.Performer = performer
        if service:
            self.Service = service
        if imagingprocedure:
            self.ImagingProcedure = imagingprocedure


class RequestedOrderDetail(OrderDetail):
    ReferringPhysician = cp.SubElementProperty([namespaces.domTag('ReferringPhysician')],
                                               valueClass=PersonReference)  # optional
    RequestingPhysician = cp.SubElementProperty([namespaces.domTag('RequestingPhysician')],
                                                valueClass=PersonReference)  # optional
    PlacerOrderNumber = cp.SubElementProperty([namespaces.domTag('PlacerOrderNumber')],
                                              valueClass=InstanceIdentifier)  # mandatory
    _props = ['ReferringPhysician', 'RequestingPhysician', 'PlacerOrderNumber']

    def __init__(self, start=None, end=None, performer=None, service=None, imagingprocedure=None,
                 referringphysician=None, requestingphysician=None, placerordernumber=None):
        """
        :param referringphysician:  a PersonReference
        :param requestingphysician: a PersonReference
        :param placerordernumber:   an InstanceIdentifier
        """
        super(RequestedOrderDetail, self).__init__(start, end, performer, service, imagingprocedure)
        self.ReferringPhysician = referringphysician
        self.RequestingPhysician = requestingphysician
        self.PlacerOrderNumber = placerordernumber


class PerformedOrderDetail(OrderDetail):
    FillerOrderNumber = cp.SubElementProperty([namespaces.domTag('FillerOrderNumber')],
                                              valueClass=InstanceIdentifier)  # optional
    ResultingClinicalInfo = cp.SubElementListProperty([namespaces.domTag('RelevantClinicalInfo')], cls=ClinicalInfo)
    _props = ['FillerOrderNumber', 'ResultingClinicalInfo']

    def __init__(self, start=None, end=None, performer=None, service=None, imagingprocedure=None,
                 fillerordernumber=None, resultingclinicalinfos=None):
        super(PerformedOrderDetail, self).__init__(start, end, performer, service, imagingprocedure)
        self.FillerOrderNumber = fillerordernumber
        if resultingclinicalinfos:
            self.ResultingClinicalInfo = resultingclinicalinfos


class WorkflowDetail(PropertyBasedPMType):
    Patient = cp.SubElementProperty([namespaces.domTag('Patient')], valueClass=PersonReference)  # optional
    AssignedLocation = cp.SubElementProperty([namespaces.domTag('AssignedLocation')],
                                             valueClass=LocationReference)  # optional
    VisitNumber = cp.SubElementProperty([namespaces.domTag('VisitNumber')], valueClass=InstanceIdentifier)  # optional
    DangerCode = cp.SubElementListProperty([namespaces.domTag('Reason')], cls=CodedValue)
    RelevantClinicalInfo = cp.SubElementListProperty([namespaces.domTag('RelevantClinicalInfo')], cls=ClinicalInfo)
    RequestedOrderDetail = cp.SubElementProperty([namespaces.domTag('RequestedOrderDetail')],
                                                 valueClass=RequestedOrderDetail)  # optional
    PerformedOrderDetail = cp.SubElementProperty([namespaces.domTag('PerformedOrderDetail')],
                                                 valueClass=PerformedOrderDetail)  # optional
    _props = ['Patient', 'AssignedLocation', 'VisitNumber', 'DangerCode',
              'RelevantClinicalInfo', 'RequestedOrderDetail', 'PerformedOrderDetail']

    def __init__(self, patient=None, assignedlocation=None, visitnumber=None, dangercode=None,
                 relevantclinicalinfo=None, requestedorderdetail=None, performedorderdetail=None):
        self.Patient = patient
        self.AssignedLocation = assignedlocation
        self.VisitNumber = visitnumber
        if dangercode:
            self.DangerCode = dangercode
        if relevantclinicalinfo:
            self.RelevantClinicalInfo = relevantclinicalinfo
        self.RequestedOrderDetail = requestedorderdetail
        self.PerformedOrderDetail = performedorderdetail


class Relation(PropertyBasedPMType):
    """Relation allows the modelling of relationships between a metric and other containtment tree entries."""
    Code = cp.SubElementProperty([namespaces.domTag('Code')], valueClass=CodedValue)  # optional
    Identification = cp.SubElementProperty([namespaces.domTag('Identification')],
                                           valueClass=InstanceIdentifier)  # optional
    Kind = cp.NodeAttributeProperty('Kind')  # required, Rcm, PS, SST, ECE, DCE, Oth
    Entries = cp.NodeAttributeListProperty('Entries')
    _props = ['Code', 'Identification', 'Kind', 'Entries']


# SafetyReq definitions
class T_Selector(PropertyBasedPMType):
    Id = cp.NodeAttributeProperty('Id')
    text = cp.NodeTextProperty()
    _props = ['Id', 'text']

    def __init__(self, id_, text):
        """
        :param id_: a string
        :param text : a string
        """
        self.Id = id_
        self.text = text

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        cls.Id.updateFromNode(obj, node)
        cls.text.updateFromNode(obj, node)
        return obj


class T_DualChannelDef(PropertyBasedPMType):
    Selector = cp.SubElementListProperty([namespaces.mdpwsTag('Selector')], cls=T_Selector)
    Algorithm = cp.NodeAttributeProperty('Algorithm')
    Transform = cp.NodeAttributeProperty('Transform')
    _props = ['Selector', 'Algorithm', 'Transform']

    def __init__(self, selectors, algorithm=None, transform=None):
        """
        :param selectors: a list of Selector objects
        :param algorithm : a string
        :param transform : a string
        """
        self.Selector = selectors
        self.Algorithm = algorithm
        self.Transform = transform

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None, None)
        cls.Selector.updateFromNode(obj, node)
        cls.Algorithm.updateFromNode(obj, node)
        cls.Transform.updateFromNode(obj, node)
        return obj


class T_SafetyContextDef(PropertyBasedPMType):
    Selector = cp.SubElementListProperty([namespaces.siTag('Selector')], cls=T_Selector)
    _props = ['Selector', ]

    def __init__(self, selectors):
        """
        :param selectors: a list of Selector objects
        """
        self.Selector = selectors

    @classmethod
    def fromNode(cls, node):
        obj = cls(None)
        cls.Selector.updateFromNode(obj, node)
        return obj


class T_SafetyReq(PropertyBasedPMType):
    DualChannelDef = cp.SubElementProperty([namespaces.siTag('DualChannelDef')],
                                           valueClass=T_DualChannelDef)  # optional
    SafetyContextDef = cp.SubElementProperty([namespaces.siTag('SafetyContextDef')],
                                             valueClass=T_SafetyContextDef)  # optional
    _props = ['DualChannelDef', 'SafetyContextDef']

    def __init__(self, dualChannelDef, safetyContextDef):
        self.DualChannelDef = dualChannelDef
        self.SafetyContextDef = safetyContextDef

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        cls.DualChannelDef.updateFromNode(obj, node)
        cls.SafetyContextDef.updateFromNode(obj, node)
        return obj


class T_Udi(PropertyBasedPMType):
    DeviceIdentifier = cp.NodeTextProperty([namespaces.domTag('DeviceIdentifier')], isOptional=False)
    HumanReadableForm = cp.NodeTextProperty([namespaces.domTag('HumanReadableForm')], isOptional=False)
    Issuer = cp.SubElementProperty([namespaces.domTag('Issuer')], valueClass=InstanceIdentifier)
    Jurisdiction = cp.SubElementProperty([namespaces.domTag('Jurisdiction')], valueClass=InstanceIdentifier)
    _props = ['DeviceIdentifier', 'HumanReadableForm', 'Issuer', 'Jurisdiction']

    def __init__(self, device_identifier=None, humanreadable_form=None, issuer=None, jurisdiction=None):
        """
        UDI fragments as defined by the FDA. (Only used in BICEPS Final)
        :param device_identifier: a string
        :param humanreadable_form: a string
        :param issuer: an InstanceIdentifier
        :param jurisdiction: an InstanceIdentifier (optional)
        """
        self.DeviceIdentifier = device_identifier
        self.HumanReadableForm = humanreadable_form
        self.Issuer = issuer
        self.Jurisdiction = jurisdiction


###################################################################################
# following : classes that serve only as name spaces


class StringEnum(str, enum.Enum):

    def __str__(self):
        return self.value


class SafetyClassification(StringEnum):
    INF = 'Inf'
    MED_A = 'MedA'
    MED_B = 'MedB'
    MED_C = 'MedC'


class MdsOperatingMode(StringEnum):
    NORMAL = 'Nml'
    DEMO = 'Dmo'
    SERVICE = 'Srv'
    MAINTENANCE = 'Mtn'


class OperatingMode(StringEnum):
    DISABLED = 'Dis'
    ENABLED = 'En'
    NA = 'NA'


class ComponentActivation(StringEnum):
    ON = 'On'
    NOT_READY = 'NotRdy'
    STANDBY = 'StndBy'
    OFF = 'Off'
    SHUTDOWN = 'Shtdn'
    FAILURE = 'Fail'


class ContextAssociation(StringEnum):
    NO_ASSOCIATION = 'No'
    PRE_ASSOCIATION = 'Pre'
    ASSOCIATED = 'Assoc'
    DISASSOCIATED = 'Dis'


class AlertConditionMonitoredLimits(StringEnum):
    ALL_ON = 'All'
    LOW_OFF = 'LoOff'
    HIGH_OFF = 'HiOff'
    ALL_OFF = 'None'


class AlertConditionPriority(StringEnum):
    NONE = 'None'
    LOW = 'Lo'
    MEDIUM = 'Me'
    HIGH = 'Hi'


class AlertConditionKind(StringEnum):
    PHYSIOLOGICAL = 'Phy'
    TECHNICAL = 'Tec'
    OTHER = 'Oth'


class AlertActivation(StringEnum):
    ON = 'On'
    OFF = 'Off'
    PAUSED = 'Psd'


class AlertSignalPresence(StringEnum):
    ON = 'On'
    OFF = 'Off'
    LATCH = 'Latch'
    ACK = 'Ack'


class AlertSignalManifestation(StringEnum):
    AUD = 'Aud'  # Aud = Audible. The ALERT SIGNAL manifests in an audible manner, i.e., the alert can be heard. Example: an alarm sound.
    VIS = 'Vis'  # Vis = Visible. The ALERT SIGNAL manifests in a visible manner, i.e., the alert can be seen. Example: a red flashing light.
    TAN = 'Tan'  # Tan = Tangible. The ALERT SIGNAL manifests in a tangible manner, i.e., the alert can be felt. Example: vibration.
    OTH = 'Oth'  # Oth = Other. The ALERT SIGNAL manifests in a manner not further specified.


class AlertSignalPrimaryLocation(StringEnum):
    LOCAL = 'Loc'
    REMOTE = 'Rem'


class MetricAvailability(StringEnum):
    INTERMITTENT = 'Intr'
    CONTINUOUS = 'Cont'


class MetricCategory(StringEnum):
    UNSPECIFIED = 'Unspec'
    MEASUREMENT = 'Msrmt'
    CALCULATION = 'Clc'
    SETTING = 'Set'
    PRESETTING = 'Preset'
    RECOMMENDATION = 'Rcmm'


class MeasurementValidity(StringEnum):
    """Level of validity of a measured value.
    Used in BICEPS Final"""
    VALID = 'Vld'
    VALIDATED_DATA = 'Vldated'
    MEASUREMENT_ONGOING = 'Ong'
    QUESTIONABLE = 'Qst'
    CALIBRATION_ONGOING = 'Calib'
    INVALID = 'Inv'
    OVERFLOW = 'Oflw'
    UNDERFLOW = 'Uflw'
    NA = 'NA'


class ChargeStatus(StringEnum):
    FUL = 'Ful'  #  Full
    CHB = 'ChB'  #  Charging Battery
    DISCHB = 'DisChB'  # Discharging Battery
    DEB = 'DEB'  # Discharged Empty Battery


class InvocationState(StringEnum):  # a namespace class
    WAIT = 'Wait'  # Wait = Waiting. The operation has been queued and waits for execution.
    START = 'Start'  # Start = Started. The execution of the operation has been started
    CANCELLED = 'Cnclld'  # Cnclld = Cancelled. The execution has been cancelled by the SERVICE PROVIDER.
    CANCELLED_MANUALLY = 'CnclldMan'  # CnclldMan = Cancelled Manually. The execution has been cancelled by the operator.
    FINISHED = 'Fin'  # Fin = Finished. The execution has been finished.
    FINISHED_MOD = 'FinMod'  # FinMod = Finished with modification. As the requested target value could not be reached, the next best value has been chosen and used as target value.
    FAILED = 'Fail'  # The execution has been failed.


class InvocationError(StringEnum):
    UNSPECIFIED = 'Unspec'  # An unspecified error has occurred. No more information about the error is available.
    UNKNOWN_OPERATION = 'Unkn'  # Unknown Operation. The HANDLE to the operation object is not known.
    INVALID_VALUE = 'Inv'  # Invalid Value. The HANDLE to the operation object does not match the invocation request message
    OTHER = 'Oth'  # Another type of error has occurred. More information on the error MAY be available.


class GenerationMode(StringEnum):
    """Describes whether METRIC data is generated by real measurements or under unreal settings (demo or test data)."""
    REAL = 'Real'  # Real Data. A value that is generated under real conditions
    TEST = 'Test'  # Test Data. A value that is arbitrary and is for testing purposes only
    DEMO = 'Demo'  # Demo Data. A value that is arbitrary and is for demonstration purposes only


class Units(object):
    UnitLess = CodedValue('262656')  # used if a metric has no unit


class DescriptionModificationTypes(StringEnum):
    CREATE = 'Crt'
    UPDATE = 'Upt'
    DELETE = 'Del'


class PatientType(StringEnum):
    UNSPECIFIED = 'Unspec'
    ADULT = 'Ad'
    ADOLESCENT = 'Ado'
    PEDIATRIC = 'Ped'
    INFANT = 'Inf'
    NEONATAL = 'Neo'
    OTHER = 'Oth'


class T_CalibrationResult(PropertyBasedPMType):
    Code = cp.SubElementProperty([namespaces.domTag('Code')], valueClass=CodedValue)
    Value = cp.SubElementProperty([namespaces.domTag('Value')], valueClass=Measurement)
    _props = ('Code', 'Value')


class T_CalibrationDocumentation(PropertyBasedPMType):
    Documentation = cp.SubElementListProperty([namespaces.domTag('Documentation')], cls=LocalizedText)
    CalibrationResult = cp.SubElementListProperty([namespaces.domTag('CalibrationResult')], cls=T_CalibrationResult)
    _props = ('Documentation', 'CalibrationResult')


class T_CalibrationState(StringEnum):
    NOT_CALIBRATED = 'No'  # No = Not Calibrated. Defines that a component is not calibrated.
    CALIBRATION_REQUIRED = 'Req'  # Req = Calibration Required. Defines that a component requires a calibration.
    RUNNING = 'Run'  # Run = Running. Defines that a calibration for a component is running.
    CALIBRATED = 'Cal'  # Cal = Calibrated. Defines that a component is calibrated.
    OTHER = 'Oth'  # Oth = Other. The calibration state is defined by other means.


class T_CalibrationType(StringEnum):
    OFFSET = 'Offset'  # Offset calibration.
    GAIN = 'Gain'  # Gain calibration
    TWO_POINT_CALIBRATION = 'TP'  # Two point calibration.
    UNSPECIFIED = 'Unspec'  # Unspecified calibration type.


class CalibrationInfo(PropertyBasedPMType):
    CalibrationDocumentation = cp.SubElementListProperty([namespaces.domTag('CalibrationDocumentation')],
                                                         cls=T_CalibrationDocumentation)
    ext_Extension = cp.ExtensionNodeProperty()
    ComponentCalibrationState = cp.NodeAttributeProperty('ComponentCalibrationState')
    Type = cp.NodeAttributeProperty('Type', impliedPyValue=T_CalibrationType.UNSPECIFIED)
    Time = cp.TimestampAttributeProperty('Time')

    _props = ('CalibrationDocumentation', 'ext_Extension', 'ComponentCalibrationState', 'Type', 'Time')



import inspect
from collections import defaultdict

from lxml import etree as etree_

from . import containerproperties as cp
from .containerbase import ContainerBase
from .. import msgtypes
from .. import observableproperties as properties
from .. import pmtypes
from ..namespaces import Prefix_Namespace as Prefix
from ..namespaces import domTag, extTag, siTag, msgTag


class AbstractDescriptorContainer(ContainerBase):
    """
    This class represents the AbstractDescriptor type. It contains a DOM node of a descriptor.
    For convenience it makes some data of that node available as members:
    nodeName: QName of the node
    nodeType: QName of the type. If no type is defined, value is None
    parentHandle: string, never None (except root node)
    handle: string, but can be None
    descriptorVersion: int
    codingSystem
    codeId
    """
    # these class variables allow easy type-checking. Derived classes will set corresponding values to True
    isSystemContextDescriptor = False
    isRealtimeSampleArrayMetricDescriptor = False
    isMetricDescriptor = False
    isOperationalDescriptor = False
    isComponentDescriptor = False
    isAlertDescriptor = False
    isAlertSignalDescriptor = False
    isAlertConditionDescriptor = False
    isContextDescriptor = False

    node = properties.ObservableProperty()  # the elementtree node

    Handle = cp.NodeAttributeProperty('Handle')
    handle = Handle
    ext_Extension = cp.ExtensionNodeProperty()
    DescriptorVersion = cp.IntegerAttributeProperty('DescriptorVersion',
                                                    defaultPyValue=0)  # optional, integer, defaults to 0
    SafetyClassification = cp.NodeAttributeProperty('SafetyClassification', impliedPyValue='Inf')  # optional
    Type = cp.SubElementProperty([domTag('Type')], valueClass=pmtypes.CodedValue)
    _props = ('Handle', 'DescriptorVersion', 'SafetyClassification', 'ext_Extension', 'Type')
    _childNodeNames = (extTag('Extension'),
                       domTag('Type'))
    NODETYPE = None
    NODENAME = None
    STATE_QNAME = None

    def __init__(self, nsmapper, nodeName, handle, parentHandle, node=None):
        super(AbstractDescriptorContainer, self).__init__(nsmapper, node=node)
        self.parentHandle = parentHandle
        if node is None:
            self.nodeName = nodeName
            self.Handle = handle
        else:
            self.nodeName = etree_.QName(node.tag)
            self._updateFromNode((node))
        self._orderedChildContainers = defaultdict(list)  # needed to keep the order,  key is node name

    @property
    def coding(self):
        return self.Type.coding if self.Type is not None else None

    @property
    def codeId(self):
        return self.Type.coding.code if self.Type is not None else None  # pylint:disable=no-member

    @property
    def codingSystem(self):
        return self.Type.coding.codingSystem if self.Type is not None else None  # pylint:disable=no-member

    def get_retrievability(self):
        """Return all retrievability data from Extension."""
        return [msgtypes.Retrievability.fromNode(x) for x in self.ext_Extension if x.tag == msgTag('Retrievability')]

    def set_retrievability(self, retrievabilities):
        """Replace all retrievability elements with provided ones in Extension.

        :param retrievabilities: Iterable of msgtypes.Retrievability."""
        for x in [x for x in self.ext_Extension if x.tag == msgTag('Retrievability')]:
            self.ext_Extension.remove(x)
        self.ext_Extension.extend([r.asEtreeNode(msgTag('Retrievability'), {}) for r in retrievabilities])

    def incrementDescriptorVersion(self):
        if self.DescriptorVersion is None:
            self.DescriptorVersion = 1
        else:
            self.DescriptorVersion += 1

    def updateDescrFromNode(self, node):
        """ This is a node msg:Descriptor"""
        # first verify that node is the same identification data as self.node
        newHandle = node.get('Handle')

        if newHandle != self.handle:
            raise ValueError('Descriptor Update Error! Old handle = {}, new handle = {}'.format(self.handle, newHandle))

        # rename new node to name of old node
        node.tag = str(self.nodeName)
        self._updateFromNode(node)
        self.node = node

    def mkNode(self, tag=None, setXsiType=False):
        my_tag = tag or self.nodeName or self.NODENAME
        return super(AbstractDescriptorContainer, self).mkNode(my_tag, setXsiType)

    def updateNode(self, setXsiType=False):
        return

    def connectChildContainers(self, node, containers):
        ret = self._connectChildNodes(node, containers)
        order = self._sortedChildNames()
        self._sortChildNodes(node, order)
        return ret

    def addChild(self, childDescriptorContainer):
        self._orderedChildContainers[childDescriptorContainer.nodeName].append(childDescriptorContainer)

    def rmChild(self, childDescriptorContainer):
        tag_specific_list = self._orderedChildContainers[childDescriptorContainer.nodeName]
        for container in tag_specific_list:
            if container.handle == childDescriptorContainer.handle:
                tag_specific_list.remove(container)

    def getOrderedChildContainers(self):
        ret = []
        for n in self._sortedChildNames():
            ret.extend(self._orderedChildContainers[n])
        return ret

    @property
    def orderedChildHandles(self):
        return [c.handle for c in self.getOrderedChildContainers()]

    def getActualValue(self, attr_name):
        """ ignores default value and implied value, e.g. returns None if value is not present in xml"""
        return getattr(self.__class__, attr_name).getActualValue(self)

    def _connectChildNodes(self, node, containers):
        ret = []
        # add all child container nodes
        for c in containers:
            n = c.mkDescriptorNode()
            node.append(n)
            ret.append((c, n))
        return ret

    def _sortedContainerProperties(self):
        """
        @return: a list of (name, object) tuples of all GenericProperties ( and subclasses)
        """
        ret = []
        classes = inspect.getmro(self.__class__)
        for cls in reversed(classes):
            try:
                names = cls.__dict__['_props']  # only access class member of this class, not parent
            except:
                continue
            for name in names:
                obj = getattr(cls, name)
                if obj is not None:
                    ret.append((name, obj))
        return ret

    def _sortedChildNames(self):
        """
        @return: a list of QNames
        """
        ret = []
        classes = inspect.getmro(self.__class__)
        for cls in reversed(classes):
            try:
                names = cls.__dict__['_childNodeNames']  # only access class member of this class, not parent
                ret.extend(names)
            except:
                continue
        return ret

    def mkDescriptorNode(self, setXsiType=True, tag=None):
        """
        Creates a lxml etree node from instance data.
        :param setXsiType:
        :param tag: tag of node, defaults to self.nodeName
        :return: an etree node
        """
        myTag = tag or self.nodeName
        node = etree_.Element(myTag, attrib={'Handle': self.handle},
                              nsmap=self.nsmapper.partialMap(Prefix.PM, Prefix.XSI))
        self._updateNode(node, setXsiType)
        order = self._sortedChildNames()
        self._sortChildNodes(node, order)
        return node

    def _sortChildNodes(self, node, ordered_tags):
        """
        raises an ValueError if a child node exist that is not listed in ordered_tags
        :param ordered_tags: a list of QNames
        """
        not_in_order = [n for n in node if n.tag not in ordered_tags]
        if len(not_in_order) > 0:
            raise ValueError('{}: not in Order:{} node={}, order={}'.format(self.__class__.__name__,
                                                                            [n.tag for n in not_in_order], node.tag,
                                                                            [o.localname for o in ordered_tags]))
        allChildNodes = node[:]
        for c in allChildNodes:
            node.remove(c)
        for qname in ordered_tags:
            for n in allChildNodes:
                if n.tag == qname:
                    node.append(n)

    def __str__(self):
        name = self.NODETYPE.localname or None
        return 'Descriptor "{}": handle={} descrVersion={} parent={}'.format(name, self.handle,
                                                                             self.DescriptorVersion, self.parentHandle)

    def __repr__(self):
        name = self.NODETYPE.localname or None
        return 'Descriptor "{}": handle={} descrVersion={} parent={}'.format(name, self.handle,
                                                                             self.DescriptorVersion, self.parentHandle)

    @classmethod
    def fromNode(cls, nsmapper, node, parentHandle):
        obj = cls(nsmapper,
                  nodeName=None,  # will be determined in constructor from node value
                  handle=None,  # will be determined in constructor from node value
                  parentHandle=parentHandle,
                  node=node)
        return obj


class AbstractDeviceComponentDescriptorContainer(AbstractDescriptorContainer):
    isComponentDescriptor = True
    ProductionSpecification = cp.SubElementListProperty([domTag('ProductionSpecification')],
                                                        cls=pmtypes.ProductionSpecification)
    _props = ('ProductionSpecification',)
    _childNodeNames = (domTag('ProductionSpecification'),)


class AbstractComplexDeviceComponentDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    _childNodeNames = (domTag('AlertSystem'),
                       domTag('Sco'),)


class MdsDescriptorContainer(AbstractComplexDeviceComponentDescriptorContainer):
    NODETYPE = domTag('MdsDescriptor')
    STATE_QNAME = domTag('MdsState')
    mNodeName = domTag('MetaData')
    Udi = cp.SubElementListProperty([mNodeName, domTag('Udi')], cls=pmtypes.T_Udi)  # 0...n
    Manufacturer = cp.SubElementListProperty([mNodeName, domTag('Manufacturer')], cls=pmtypes.LocalizedText)
    ModelName = cp.SubElementListProperty([mNodeName, domTag('ModelName')], cls=pmtypes.LocalizedText)
    ModelNumber = cp.NodeTextProperty([mNodeName, domTag('ModelNumber')])
    SerialNumber = cp.SubElementListProperty([mNodeName, domTag('SerialNumber')],
                                             cls=pmtypes.ElementWithTextOnly)
    _props = ('Udi', 'Manufacturer', 'ModelName', 'ModelNumber', 'SerialNumber')
    _childNodeNames = (domTag('MetaData'),
                       domTag('SystemContext'),
                       domTag('Clock'),
                       domTag('Battery'),
                       domTag('ApprovedJurisdictions'),
                       domTag('Vmd'),
                       )

    def connectChildContainers(self, node, containers):
        ret = super(MdsDescriptorContainer, self).connectChildContainers(node, containers)
        self._sortMetaData(node)
        return ret

    def _sortMetaData(self, node):
        # sort also MetaData SubNode
        metaDataNode = node.find(domTag('MetaData'))
        if metaDataNode is not None:
            order = (domTag('Udi'),
                     domTag('Manufacturer'),
                     domTag('ModelName'),
                     domTag('ModelNumber'),
                     domTag('SerialNumber'),
                     )
            self._sortChildNodes(metaDataNode, order)

    def mkDescriptorNode(self, setXsiType=True, tag=None):
        """returns a node without any child with a handle"""
        node = super(MdsDescriptorContainer, self).mkDescriptorNode(setXsiType, tag)
        self._sortMetaData(node)
        return node


class VmdDescriptorContainer(AbstractComplexDeviceComponentDescriptorContainer):
    NODETYPE = domTag('VmdDescriptor')
    STATE_QNAME = domTag('VmdState')
    _childNodeNames = (domTag('Channel'),)


class ChannelDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    NODETYPE = domTag('ChannelDescriptor')
    STATE_QNAME = domTag('ChannelState')
    _childNodeNames = (domTag('Metric'),)


class ClockDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    NODETYPE = domTag('ClockDescriptor')
    STATE_QNAME = domTag('ClockState')
    TimeProtocol = cp.SubElementListProperty([domTag('TimeProtocol')], cls=pmtypes.CodedValue)
    Resolution = cp.DurationAttributeProperty('Resolution')  # optional,  xsd:duration
    _props = ('TimeProtocol', 'Resolution')
    _childNodeNames = (domTag('TimeProtocol'),)


class BatteryDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    NODETYPE = domTag('BatteryDescriptor')
    STATE_QNAME = domTag('BatteryState')
    CapacityFullCharge = cp.SubElementProperty([domTag('CapacityFullCharge')],
                                               valueClass=pmtypes.Measurement)  # optional
    CapacitySpecified = cp.SubElementProperty([domTag('CapacitySpecified')], valueClass=pmtypes.Measurement)  # optional
    VoltageSpecified = cp.SubElementProperty([domTag('VoltageSpecified')], valueClass=pmtypes.Measurement)  # optional
    _props = ('CapacityFullCharge', 'CapacitySpecified', 'VoltageSpecified')
    _childNodeNames = (domTag('CapacityFullCharge'),
                       domTag('CapacitySpecified'),
                       domTag('VoltageSpecified'),
                       )


class ScoDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    NODETYPE = domTag('ScoDescriptor')
    STATE_QNAME = domTag('ScoState')
    # This has AbstractOperationDescriptor children. Not modeled here
    _childNodeNames = (domTag('Operation'),
                       )


class AbstractMetricDescriptorContainer(AbstractDescriptorContainer):
    isMetricDescriptor = True
    Unit = cp.SubElementProperty([domTag('Unit')], valueClass=pmtypes.CodedValue)
    BodySite = cp.SubElementListProperty([domTag('BodySite')], cls=pmtypes.CodedValue)
    Relation = cp.SubElementListProperty([domTag('Relation')], cls=pmtypes.Relation)  # o...n
    MetricCategory = cp.NodeAttributeProperty('MetricCategory', defaultPyValue='Unspec')  # required
    DerivationMethod = cp.NodeAttributeProperty('DerivationMethod')  # optional
    #  There is an implied value defined, but it is complicated, therefore here not implemented:
    # - If pm:AbstractDescriptor/@MetricCategory is "Set" or "Preset", then the default value of DerivationMethod is "Man"
    # - If pm:AbstractDescriptor/@MetricCategory is "Clc", "Msrmt", "Rcmm", then the default value of DerivationMethod is "Auto"
    # - If pm:AbstractDescriptor/@MetricCategory is "Unspec", then no default value is being implied</xsd:documentation>
    MetricAvailability = cp.NodeAttributeProperty('MetricAvailability', defaultPyValue='Cont')  # required
    MaxMeasurementTime = cp.DurationAttributeProperty('MaxMeasurementTime')  # optional,  xsd:duration
    MaxDelayTime = cp.DurationAttributeProperty('MaxDelayTime')  # optional,  xsd:duration
    DeterminationPeriod = cp.DurationAttributeProperty('DeterminationPeriod')  # optional,  xsd:duration
    LifeTimePeriod = cp.DurationAttributeProperty('LifeTimePeriod')  # optional,  xsd:duration
    ActivationDuration = cp.DurationAttributeProperty('ActivationDuration')  # optional,  xsd:duration
    _props = (
        'Unit', 'BodySite', 'Relation', 'MetricCategory', 'DerivationMethod', 'MetricAvailability',
        'MaxMeasurementTime', 'MaxDelayTime', 'DeterminationPeriod', 'LifeTimePeriod', 'ActivationDuration')
    _childNodeNames = (domTag('Unit'),
                       domTag('BodySite'),
                       domTag('Relation'),
                       )

    def addChild(self, childDescriptorContainer):
        raise ValueError('Metric can not have children')

    def rmChild(self, childDescriptorContainer):
        raise ValueError('Metric can not have children')


class NumericMetricDescriptorContainer(AbstractMetricDescriptorContainer):
    NODETYPE = domTag('NumericMetricDescriptor')
    STATE_QNAME = domTag('NumericMetricState')
    TechnicalRange = cp.SubElementListProperty([domTag('TechnicalRange')], cls=pmtypes.Range)
    Resolution = cp.DecimalAttributeProperty('Resolution')  # optional
    AveragingPeriod = cp.DurationAttributeProperty('AveragingPeriod')  # optional
    _props = ('TechnicalRange', 'Resolution', 'AveragingPeriod')
    _childNodeNames = (domTag('TechnicalRange'),
                       )


class StringMetricDescriptorContainer(AbstractMetricDescriptorContainer):
    NODETYPE = domTag('StringMetricDescriptor')
    STATE_QNAME = domTag('StringMetricState')


class EnumStringMetricDescriptorContainer(StringMetricDescriptorContainer):
    NODETYPE = domTag('EnumStringMetricDescriptor')
    STATE_QNAME = domTag('EnumStringMetricState')
    AllowedValue = cp.SubElementListProperty([domTag('AllowedValue')], cls=pmtypes.AllowedValue)
    _props = ('AllowedValue',)
    _childNodeNames = (domTag('AllowedValue'),
                       )


class RealTimeSampleArrayMetricDescriptorContainer(AbstractMetricDescriptorContainer):
    isRealtimeSampleArrayMetricDescriptor = True
    NODETYPE = domTag('RealTimeSampleArrayMetricDescriptor')
    STATE_QNAME = domTag('RealTimeSampleArrayMetricState')
    TechnicalRange = cp.SubElementListProperty([domTag('TechnicalRange')], cls=pmtypes.Range)
    Resolution = cp.DecimalAttributeProperty('Resolution')  # optional
    SamplePeriod = cp.DurationAttributeProperty('SamplePeriod')  # required
    _props = ('TechnicalRange', 'Resolution', 'SamplePeriod')
    _childNodeNames = (domTag('TechnicalRange'),
                       )


class DistributionSampleArrayMetricDescriptorContainer(AbstractMetricDescriptorContainer):
    NODETYPE = domTag('DistributionSampleArrayMetricDescriptor')
    STATE_QNAME = domTag('DistributionSampleArrayMetricState')
    TechnicalRange = cp.SubElementListProperty([domTag('TechnicalRange')], cls=pmtypes.Range)
    DomainUnit = cp.SubElementProperty([domTag('DomainUnit')], valueClass=pmtypes.CodedValue)
    DistributionRange = cp.SubElementProperty([domTag('DistributionRange')], valueClass=pmtypes.Range)
    Resolution = cp.DecimalAttributeProperty('Resolution')  # required
    _props = ('TechnicalRange', 'DomainUnit', 'DistributionRange', 'Resolution')
    _childNodeNames = (domTag('TechnicalRange'), domTag('DomainUnit'), domTag('DistributionRange'))


class AbstractOperationDescriptorContainer(AbstractDescriptorContainer):
    isOperationalDescriptor = True
    OperationTarget = cp.NodeAttributeProperty('OperationTarget')
    SafetyReq = cp.SubElementProperty([extTag('Extension'), siTag('SafetyReq')], valueClass=pmtypes.T_SafetyReq)
    InvocationEffectiveTimeout = cp.DurationAttributeProperty('InvocationEffectiveTimeout')  # optional  xsd:duration
    MaxTimeToFinish = cp.DurationAttributeProperty('MaxTimeToFinish')  # optional  xsd:duration
    Retriggerable = cp.BooleanAttributeProperty('Retriggerable', impliedPyValue=True)  # optional
    #  AccessLevel can be: Usr (User), CSUsr (Clinical Super User), RO (Responsible Organization),
    #  SP (Service Personnel), Oth (Other)
    AccessLevel = cp.NodeAttributeProperty('AccessLevel', impliedPyValue='Usr')
    _props = (
        'OperationTarget', 'SafetyReq', 'InvocationEffectiveTimeout', 'MaxTimeToFinish', 'Retriggerable', 'AccessLevel')


class SetValueOperationDescriptorContainer(AbstractOperationDescriptorContainer):
    NODETYPE = domTag('SetValueOperationDescriptor')
    STATE_QNAME = domTag('SetValueOperationState')


class SetStringOperationDescriptorContainer(AbstractOperationDescriptorContainer):
    NODETYPE = domTag('SetStringOperationDescriptor')
    STATE_QNAME = domTag('SetStringOperationState')
    MaxLength = cp.IntegerAttributeProperty('MaxLength')
    _props = ('MaxLength',)


class AbstractSetStateOperationDescriptor(AbstractOperationDescriptorContainer):
    ModifiableData = cp.SubElementListProperty([domTag('ModifiableData')], cls=pmtypes.ElementWithTextOnly)
    _props = ('ModifiableData',)
    _childNodeNames = (domTag('ModifiableData'),)


class SetContextStateOperationDescriptorContainer(AbstractSetStateOperationDescriptor):
    NODETYPE = domTag('SetContextStateOperationDescriptor')
    STATE_QNAME = domTag('SetContextStateOperationState')


class SetMetricStateOperationDescriptorContainer(AbstractSetStateOperationDescriptor):
    NODETYPE = domTag('SetMetricStateOperationDescriptor')
    STATE_QNAME = domTag('SetMetricStateOperationState')


class SetComponentStateOperationDescriptorContainer(AbstractSetStateOperationDescriptor):
    NODETYPE = domTag('SetComponentStateOperationDescriptor')
    STATE_QNAME = domTag('SetComponentStateOperationState')


class SetAlertStateOperationDescriptorContainer(AbstractSetStateOperationDescriptor):
    NODETYPE = domTag('SetAlertStateOperationDescriptor')
    STATE_QNAME = domTag('SetAlertStateOperationState')


class ActivateOperationDescriptorContainer(AbstractSetStateOperationDescriptor):
    NODETYPE = domTag('ActivateOperationDescriptor')
    STATE_QNAME = domTag('ActivateOperationState')
    Argument = cp.SubElementListProperty([domTag('Argument')], cls=pmtypes.Argument)
    _props = ('Argument',)
    _childNodeNames = (domTag('Argument'),)


class AbstractAlertDescriptorContainer(AbstractDescriptorContainer):
    """AbstractAlertDescriptor acts as a base class for all alert descriptors that contain static alert meta information.
     This class has nor specific data."""
    isAlertDescriptor = True


class AlertSystemDescriptorContainer(AbstractAlertDescriptorContainer):
    """AlertSystemDescriptor describes an ALERT SYSTEM to detect ALERT CONDITIONs and generate ALERT SIGNALs, 
    which belong to specific ALERT CONDITIONs.
    ALERT CONDITIONs are represented by a list of pm:AlertConditionDescriptor ELEMENTs and 
    ALERT SIGNALs are represented by a list of pm:AlertSignalDescriptor ELEMENTs.
    """
    NODETYPE = domTag('AlertSystemDescriptor')
    STATE_QNAME = domTag('AlertSystemState')
    MaxPhysiologicalParallelAlarms = cp.IntegerAttributeProperty('MaxPhysiologicalParallelAlarms')
    MaxTechnicalParallelAlarms = cp.IntegerAttributeProperty('MaxTechnicalParallelAlarms')
    SelfCheckPeriod = cp.DurationAttributeProperty('SelfCheckPeriod')
    _props = ('MaxPhysiologicalParallelAlarms', 'MaxTechnicalParallelAlarms', 'SelfCheckPeriod')
    _childNodeNames = (domTag('AlertCondition'),
                       domTag('AlertSignal'))


class AlertConditionDescriptorContainer(AbstractAlertDescriptorContainer):
    """An ALERT CONDITION contains the information about a potentially or actually HAZARDOUS SITUATION.
      Examples: a physiological alarm limit has been exceeded or a sensor has been unplugged."""
    isAlertConditionDescriptor = True
    NODETYPE = domTag('AlertConditionDescriptor')
    STATE_QNAME = domTag('AlertConditionState')
    Source = cp.SubElementListProperty([domTag('Source')], # a list of 0...n pm:HandleRef elements
                                       cls=pmtypes.ElementWithTextOnly)
    CauseInfo = cp.SubElementListProperty([domTag('CauseInfo')], # a list of 0...n pm:CauseInfo elements
                                          cls=pmtypes.CauseInfo)
    Kind = cp.NodeAttributeProperty('Kind',  # required, type=AlertConditionKind
                                    defaultPyValue=pmtypes.AlertConditionKind.OTHER)
    Priority = cp.NodeAttributeProperty('Priority',  # required, type= AlertConditionPriority
                                        defaultPyValue=pmtypes.AlertConditionPriority.NONE)
    DefaultConditionGenerationDelay = cp.DurationAttributeProperty('DefaultConditionGenerationDelay',
                                                                   impliedPyValue=0)  # optional
    CanEscalate = cp.NodeAttributeProperty('CanEscalate')  # AlertConditionPriority, without 'None'
    CanDeescalate = cp.NodeAttributeProperty('CanDeescalate')  # AlertConditionPriority, without 'Hi'
    _props = (
        'Source', 'CauseInfo', 'Kind', 'Priority', 'DefaultConditionGenerationDelay', 'CanEscalate', 'CanDeescalate')
    _childNodeNames = (domTag('Source'),
                       domTag('CauseInfo'))


class LimitAlertConditionDescriptorContainer(AlertConditionDescriptorContainer):
    NODETYPE = domTag('LimitAlertConditionDescriptor')
    STATE_QNAME = domTag('LimitAlertConditionState')
    MaxLimits = cp.SubElementProperty([domTag('MaxLimits')], valueClass=pmtypes.Range)
    AutoLimitSupported = cp.BooleanAttributeProperty('AutoLimitSupported', impliedPyValue=False)
    _props = ('MaxLimits', 'AutoLimitSupported',)
    _childNodeNames = (domTag('MaxLimits'),)


class AlertSignalDescriptorContainer(AbstractAlertDescriptorContainer):
    isAlertSignalDescriptor = True
    NODETYPE = domTag('AlertSignalDescriptor')
    STATE_QNAME = domTag('AlertSignalState')
    ConditionSignaled = cp.NodeAttributeProperty('ConditionSignaled')  # required, a HandleRef
    Manifestation = cp.NodeAttributeProperty(
        'Manifestation')  # required, an AlertSignalManifestation ('Aud', 'Vis', 'Tan', 'Oth')
    Latching = cp.BooleanAttributeProperty('Latching')  # required
    DefaultSignalGenerationDelay = cp.DurationAttributeProperty('DefaultSignalGenerationDelay',
                                                                impliedPyValue=0)  # optional, defaults to 0 seconds
    SignalDelegationSupported = cp.BooleanAttributeProperty('SignalDelegationSupported',
                                                            impliedPyValue=False)  # optional, defaults to false
    AcknowledgementSupported = cp.BooleanAttributeProperty('AcknowledgementSupported',
                                                           impliedPyValue=False)  # optional, defaults to false
    AcknowledgeTimeout = cp.DurationAttributeProperty('AcknowledgeTimeout')  # optional
    _props = ('ConditionSignaled', 'Manifestation', 'Latching', 'DefaultSignalGenerationDelay',
              'SignalDelegationSupported', 'AcknowledgementSupported', 'AcknowledgeTimeout')


class SystemContextDescriptorContainer(AbstractDeviceComponentDescriptorContainer):
    isSystemContextDescriptor = True
    NODETYPE = domTag('SystemContextDescriptor')
    STATE_QNAME = domTag('SystemContextState')
    # Child elements are not modeled here
    _childNodeNames = (domTag('PatientContext'),
                       domTag('LocationContext'),
                       domTag('EnsembleContext'),
                       domTag('OperatorContext'),
                       domTag('WorkflowContext'),
                       domTag('MeansContext'),
                       )


class AbstractContextDescriptorContainer(AbstractDescriptorContainer):
    isContextDescriptor = True


class PatientContextDescriptorContainer(AbstractContextDescriptorContainer):
    NODETYPE = domTag('PatientContextDescriptor')
    STATE_QNAME = domTag('PatientContextState')


class LocationContextDescriptorContainer(AbstractContextDescriptorContainer):
    NODETYPE = domTag('LocationContextDescriptor')
    STATE_QNAME = domTag('LocationContextState')


class WorkflowContextDescriptorContainer(AbstractContextDescriptorContainer):
    NODETYPE = domTag('WorkflowContextDescriptor')
    STATE_QNAME = domTag('WorkflowContextState')


class OperatorContextDescriptorContainer(AbstractContextDescriptorContainer):
    NODETYPE = domTag('OperatorContextDescriptor')
    STATE_QNAME = domTag('OperatorContextState')


class MeansContextDescriptorContainer(AbstractContextDescriptorContainer):
    NODETYPE = domTag('MeansContextDescriptor')
    STATE_QNAME = domTag('MeansContextState')


class EnsembleContextDescriptorContainer(AbstractContextDescriptorContainer):
    NODETYPE = domTag('EnsembleContextDescriptor')
    STATE_QNAME = domTag('EnsembleContextState')


_name_class_lookup = {
    domTag('Battery'): BatteryDescriptorContainer,
    domTag('BatteryDescriptor'): BatteryDescriptorContainer,
    domTag('Mds'): MdsDescriptorContainer,
    domTag('MdsDescriptor'): MdsDescriptorContainer,
    domTag('Vmd'): VmdDescriptorContainer,
    domTag('VmdDescriptor'): VmdDescriptorContainer,
    domTag('Sco'): ScoDescriptorContainer,
    domTag('ScoDescriptor'): ScoDescriptorContainer,
    domTag('Channel'): ChannelDescriptorContainer,
    domTag('ChannelDescriptor'): ChannelDescriptorContainer,
    domTag('Clock'): ClockDescriptorContainer,
    domTag('ClockDescriptor'): ClockDescriptorContainer,
    domTag('SystemContext'): SystemContextDescriptorContainer,
    domTag('SystemContextDescriptor'): SystemContextDescriptorContainer,
    domTag('PatientContext'): PatientContextDescriptorContainer,
    domTag('LocationContext'): LocationContextDescriptorContainer,
    domTag('PatientContextDescriptor'): PatientContextDescriptorContainer,
    domTag('LocationContextDescriptor'): LocationContextDescriptorContainer,
    domTag('WorkflowContext'): WorkflowContextDescriptorContainer,
    domTag('WorkflowContextDescriptor'): WorkflowContextDescriptorContainer,
    domTag('OperatorContext'): OperatorContextDescriptorContainer,
    domTag('OperatorContextDescriptor'): OperatorContextDescriptorContainer,
    domTag('MeansContext'): MeansContextDescriptorContainer,
    domTag('MeansContextDescriptor'): MeansContextDescriptorContainer,
    domTag('EnsembleContext'): EnsembleContextDescriptorContainer,
    domTag('EnsembleContextDescriptor'): EnsembleContextDescriptorContainer,
    domTag('AlertSystem'): AlertSystemDescriptorContainer,
    domTag('AlertSystemDescriptor'): AlertSystemDescriptorContainer,
    domTag('AlertCondition'): AlertConditionDescriptorContainer,
    domTag('AlertConditionDescriptor'): AlertConditionDescriptorContainer,
    domTag('AlertSignal'): AlertSignalDescriptorContainer,
    domTag('AlertSignalDescriptor'): AlertSignalDescriptorContainer,
    domTag('StringMetricDescriptor'): StringMetricDescriptorContainer,
    domTag('EnumStringMetricDescriptor'): EnumStringMetricDescriptorContainer,
    domTag('NumericMetricDescriptor'): NumericMetricDescriptorContainer,
    domTag('RealTimeSampleArrayMetricDescriptor'): RealTimeSampleArrayMetricDescriptorContainer,
    domTag('DistributionSampleArrayMetricDescriptor'): DistributionSampleArrayMetricDescriptorContainer,
    domTag('LimitAlertConditionDescriptor'): LimitAlertConditionDescriptorContainer,
    domTag('SetValueOperationDescriptor'): SetValueOperationDescriptorContainer,
    domTag('SetStringOperationDescriptor'): SetStringOperationDescriptorContainer,
    domTag('ActivateOperationDescriptor'): ActivateOperationDescriptorContainer,
    domTag('SetContextStateOperationDescriptor'): SetContextStateOperationDescriptorContainer,
    domTag('SetMetricStateOperationDescriptor'): SetMetricStateOperationDescriptorContainer,
    domTag('SetComponentStateOperationDescriptor'): SetComponentStateOperationDescriptorContainer,
    domTag('SetAlertStateOperationDescriptor'): SetAlertStateOperationDescriptorContainer,
}


def getContainerClass(qNameType):
    """
    :param qNameType: a QName instance
    """
    # first check type, this is more specific. 
    return _name_class_lookup.get(qNameType)



import base64
import hashlib
from lxml import etree as etree_
from sdc11073.namespaces import Prefix_Namespace as Prefix
from sdc11073.namespaces import mdpwsTag


def B64Sha1(valueString):
    """Gets standard base64 coded sha1 sum of input file."""
    # pylint: disable=E1101
    if isinstance(valueString, str):
        valueString = valueString.encode('utf-8')
    return base64.standard_b64encode(hashlib.sha1(valueString).digest()).decode()

def Sha1(valueString):
    """Gets standard base64 coded sha1 sum of input file."""
    # pylint: disable=E1101
    if isinstance(valueString, str):
        valueString = valueString.encode('utf-8')
    return hashlib.sha1(valueString).hexdigest()


class SafetyInfoHeader(object):
    def __init__(self, dualChannelValues, safetyContextValues, algorithm=None):
        self.dualChannelValues = dualChannelValues
        self.safetyContextValues = safetyContextValues
        self._algorithm = algorithm


    def _asEtreeNode(self):
        safetyInfo = etree_.Element(mdpwsTag('SafetyInfo'), nsmap=Prefix.partialMap(Prefix.MDPWS))
        if self.dualChannelValues:
            dualChannel = etree_.SubElement(safetyInfo, mdpwsTag('DualChannel'))
            for ref, value in self.dualChannelValues.items():
                dcValue = etree_.SubElement(dualChannel, mdpwsTag('DcValue'))
                dcValue.set('ReferencedSelector', ref)
                algorithm = self._algorithm or Sha1
                dcValue.text = algorithm(value)
        
        if self.safetyContextValues:
            safetyContext = etree_.SubElement(safetyInfo, mdpwsTag('SafetyContext'))
            for ref, value in self.safetyContextValues.items():
                ctxtValue = etree_.SubElement(safetyContext, mdpwsTag('CtxtValue'))
                ctxtValue.set('ReferencedSelector', ref)
                ctxtValue.text = value
        return safetyInfo


    def asEtreeSubNode(self, rootNode):
        rootNode.append(self._asEtreeNode())

    @classmethod
    def fromEtreeNode(cls, rootNode): #pylint: disable=unused-argument
        raise NotImplementedError



class _Selector(object):
    def __init__(self, xpathString):
        self.xpathString = xpathString



class DualChannelDef(object):
    """ Definition is located in MdDescription"""
    def __init__(self, algorithm, transform, selectorDict):
        self.algorithm = algorithm
        self.transform = transform
        self.selectorDict = selectorDict
    
    
    @classmethod
    def fromEtreeNode(cls, node):
        algorithm = node.get('Algorithm')
        transform = node.get('Transform')
        selectorDict = {}
        for s in node.findall(mdpwsTag('Selector')):
            id_ = s.get('Id')
            text = s.text
            selectorDict[id_] = _Selector(text)
        
        return cls(algorithm, transform, selectorDict)
        


import socket
import threading
import time
import traceback
from http.server import HTTPServer
from urllib.parse import urlparse

from .exceptions import HTTPRequestHandlingError, InvalidPathError, InvalidActionError
from .. import commlog
from .. import loghelper
from .. import pysoap
from ..httprequesthandler import HTTPRequestHandler, mkchunks


class MyThreadingMixIn(object):

    def process_request_thread(self, request, client_address):
        """Same as in BaseServer but as a thread.

        In addition, exception handling is done here.

        """
        try:
            self.finish_request(request, client_address)
            self.shutdown_request(request)
        except Exception as ex:
            if self.dispatcher is not None:
                # only 
                self.handle_error(request, client_address)
            self.shutdown_request(request)


    def process_request(self, request, client_address):
        """Start a new thread to process the request."""
        t = threading.Thread(target = self.process_request_thread,
                             args = (request, client_address),
                             name='SubscrRecv{}'.format(client_address))
        t.daemon = True
        t.start()
        self.threads.append((t,request, client_address))


class ThreadingHTTPServer(MyThreadingMixIn, HTTPServer):
    """ Each request is handled in a thread.
    Following recipe from https://pymotw.com/2/BaseHTTPServer/index.html#module-BaseHTTPServer
    """
    def __init__(self, *args, **kwargs):
        HTTPServer. __init__(self, *args, **kwargs)
        self.daemon_threads = True
        self.threads = []
        self.dispatcher = None

class DevicesDispatcher(object):
    """ Dispatch to one of the registered devices, based on url"""
    def __init__(self, logger):
        self._logger = logger
        self.deviceByUrl = {}       # lookup for requests

    def register_device_dispatcher(self, path, dispatcher):
        if path.startswith('/'):
            path = path[1:]
        if path.endswith('/'):
            path = path[:-1]
        if path in self.deviceByUrl:
            raise RuntimeError('Path "{}" already registered'.format(path))
        self.deviceByUrl[path] = dispatcher

    def get_device_dispatcher(self, path):
        _path = path[1:] if path.startswith('/') else path
        for url, dispatcher in self.deviceByUrl.items():
            if _path.startswith(url):
                return dispatcher
        raise HTTPRequestHandlingError(status=404, reason='not found', soapfault=b'client error')

    def on_post(self, path, headers, request):
        return self.get_device_dispatcher(path).on_post(path, headers, request)

    def on_get(self, path, headers):
        return self.get_device_dispatcher(path).on_get(path, headers)


class HostedServiceDispatcher(object):
    """ receiver of all messages"""

    def __init__(self, sdc_definitions, logger):
        self.sdc_definitions = sdc_definitions
        self._logger = logger
        self.hostedServiceByUrl = {}       # lookup for requests
        self.hostedServices = []

    def register_hosted_service(self, hosted_service):
        path = hosted_service.epr
        if path.endswith('/'):
            path = path[:-1]
        if path in self.hostedServiceByUrl:
            raise RuntimeError('Path "{}" already registered'.format(path))
        self.hostedServiceByUrl[path] = hosted_service
        self.hostedServices.append(hosted_service)

    def on_post(self, path, headers, request):
        """Method converts the http request into a soap envelope and calls dispatchSoapRequest.
           Return of dispatchSoapRequest (soap envelope) is converted back to a string."""
        commlog.defaultLogger.logSoapReqIn(request, 'POST')
        normalizedRequest = self.sdc_definitions.normalizeXMLText(request)
        # execute the method
        soapEnvelope = pysoap.soapenvelope.ReceivedSoap12Envelope.fromXMLString(normalizedRequest)
        response = self._dispatchSoapRequest(path, headers, soapEnvelope)
        normalized_response_xml_string = response.as_xml()
        return self.sdc_definitions.denormalizeXMLText(normalized_response_xml_string)

    def _dispatchSoapRequest(self, path, header, soapEnvelope):
        # path is a string like /0105a018-8f4c-4199-9b04-aff4835fd8e9/StateEvent, without http:/servername:port
        hostedService = self.hostedServiceByUrl.get(path)
        if not hostedService:
            raise InvalidPathError(soapEnvelope, path)
        try:
            return hostedService.dispatchSoapRequest( path, header, soapEnvelope)
        except InvalidActionError as ex:
            # error: no handler for this action; log this error with all known paths, then re-raise
            all_actions = []
            for dispatcher in self.hostedServices:
                all_actions.extend(', '.join([dispatcher.epr, k]) for k in dispatcher.getActions())

            txt = 'HostedServiceDispatcher.dispatchSoapRequest: {} , known=\n{}'.format(ex, '\n'.join(all_actions))
            self._logger.error(txt)
            raise

    def on_get(self, path, httpHeaders):
        """ Get Requests are handled as they are, no soap envelopes"""
        response_string = self._dispatchGetRequest(path, httpHeaders)
        return  self.sdc_definitions.denormalizeXMLText(response_string)

    def _dispatchGetRequest(self, path, httpHeaders):
        parsedPath = urlparse(path)
        p = parsedPath.path
        if p.endswith('/'):
            p = p[:-1]
        dispatcher = self.hostedServiceByUrl.get(p)
        if dispatcher is None:
            raise KeyError(
                'HostedServiceDispatcher.dispatchGetRequest: unknown path "{}", known = {}'.format(p, self.hostedServiceByUrl.keys()))
        response_string = dispatcher.dispatchGetRequest(parsedPath, httpHeaders)
        return  self.sdc_definitions.denormalizeXMLText(response_string)


class _SdcServerRequestHandler(HTTPRequestHandler):
    protocol_version = "HTTP/1.1"  # this enables keep-alive
    # This server does NOT disable nagle algorithm. It sends Large responses,
    # and network efficiency is more important than short latencies.
    disable_nagle_algorithm = False

    def do_POST(self):
        """SOAP POST gateway"""
        try:
            devices_dispatcher = self.server.dispatcher
            if devices_dispatcher is None:
                # close this connection
                self.close_connection = True
                response_xml_string = b'received a POST request, but have no dispatcher'
                self.send_response(404)  # not found
            else:
                request = self._read_request()
                commlog.defaultLogger.logSoapReqIn(request, 'POST')
                try:
                    #delegate handling to on_post method of dispatcher
                    response_xml_string = devices_dispatcher.on_post(self.path, self.headers, request)
                    http_status = 200
                    http_reason = 'Ok'
                    # MDPWS:R0007 A text SOAP envelope shall be serialized using utf-8 character encoding
                    assert (b'utf-8' in response_xml_string[:100].lower())
                except HTTPRequestHandlingError as ex:
                    response_xml_string = ex.soapfault
                    http_status = ex.status
                    http_reason = ex.reason

                commlog.defaultLogger.logSoapRespOut(response_xml_string, 'POST')
                self.send_response(http_status, http_reason)
            response_xml_string = self._compressIfRequired(response_xml_string)
            self.send_header("Content-type", "application/soap+xml; charset=utf-8")
            if self.server.chunked_response:
                self.send_header("transfer-encoding", "chunked")
                self.end_headers()
                self.wfile.write(mkchunks(response_xml_string))
            else:
                self.send_header("Content-length", str(len(response_xml_string)))
                self.end_headers()
                self.wfile.write(response_xml_string)
        except Exception as ex:
            # make an error 500 response with the soap fault as content
            self.server.logger.error(traceback.format_exc())
            # we must create a soapEnvelope in order to generate a SoapFault
            dev_dispatcher = devices_dispatcher.get_device_dispatcher(self.path)
            normalizedRequest = dev_dispatcher.sdc_definitions.normalizeXMLText(request)
            soapEnvelope = pysoap.soapenvelope.ReceivedSoap12Envelope.fromXMLString(normalizedRequest)

            response = pysoap.soapenvelope.SoapFault(soapEnvelope, code=pysoap.soapenvelope.SoapFaultCode.SENDER,
                                                          reason=str(ex))
            normalized_response_xml_string = response.as_xml()
            response_xml_string = dev_dispatcher.sdc_definitions.denormalizeXMLText(normalized_response_xml_string)
            self.send_response(500)
            self.send_header("Content-type", "application/soap+xml; charset=utf-8")
            self.send_header("Content-length", str(len(response_xml_string)))
            self.end_headers()
            self.wfile.write(response_xml_string)

    def do_GET(self):
        parsedPath = urlparse(self.path)
        try:
            commlog.defaultLogger.logSoapReqIn(b'', 'GET')  # GET has no content, log it to document duration of processing
            response_string = self.server.dispatcher.on_get(self.path, self.headers)
            self.send_response(200, 'Ok')
            response_string = self._compressIfRequired(response_string)
            commlog.defaultLogger.logSoapRespOut(response_string, 'GET')
            if parsedPath.query == 'wsdl':
                content_type = "text/xml; charset=utf-8"
            else:
                content_type = "application/soap+xml; charset=utf-8"
        except Exception as ex:
            self.send_response(500)
            response_string = str(ex)
            content_type = "text"

        self.send_header("Content-Type", content_type)
        self.send_header("Content-Length", str(len(response_string)))
        self.end_headers()
        self.wfile.write(response_string)


class HttpServerThread(threading.Thread):
    
    def __init__(self, my_ipaddress, sslContext, supportedEncodings, log_prefix=None, chunked_responses=False):
        """
        :param my_ipaddress:
        :param sslContext:
        :param supportedEncodings: a list od strings
        """
        super(HttpServerThread, self).__init__(name='Dev_SdcHttpServerThread')
        self.daemon = True

        self._my_ipaddress = my_ipaddress
        self._sslContext = sslContext
        self.my_port = None
        self.httpd = None
        self.supportedEncodings = supportedEncodings

        self._logger = loghelper.getLoggerAdapter('sdc.device.httpsrv', log_prefix)
        self.chunked_responses = chunked_responses
        # create and set up the dispatcher for all incoming requests
        self.devices_dispatcher = DevicesDispatcher(self._logger)
        self.started_evt = threading.Event() # helps to wait until thread has initialised is variables


    def run(self):
        try:
            myport = 0  # zero means that OS selects a free port
            self.httpd = ThreadingHTTPServer((self._my_ipaddress, myport), _SdcServerRequestHandler)
            self.httpd.chunked_response = self.chunked_responses # monkey-patching, value needed by _SdcServerRequestHandler
            # add use compression flag to the server
            setattr(self.httpd, 'supportedEncodings', self.supportedEncodings)
            self.httpd.logger = self._logger # add logger by monkey-pathing
            if self._sslContext is not None:
                self.httpd.socket = self._sslContext.wrap_socket(self.httpd.socket,
                                                                 server_side=True,
                                                                 do_handshake_on_connect=False)
            self.my_port = self.httpd.server_port
            self.httpd.dispatcher = self.devices_dispatcher

            self.started_evt.set()
            self.httpd.serve_forever()
        except Exception:
            self._logger.error('Unhandled Exception at thread runtime. Thread will abort! {}'.format(traceback.format_exc()))
            raise

    def setCompressionFlag(self, useCompression):
        """Sets use compression attribute on the http server to be used in handler
        :param useCompression: bool flag 
        """
        self.httpd.useCompression = useCompression
   
    def stop(self, closeAllConnections=True):
        self.httpd.shutdown()
        self.join(timeout=5)
        self.httpd.socket.close()
        if closeAllConnections:
            if self.httpd.dispatcher is not None:
                self.httpd.dispatcher = None # this leads to a '503' reaction in SOAPNotificationsHandler
            for thr in self.httpd.threads:
                thread, request, client_addr = thr
                if thread.is_alive():
                    try:
                        request.shutdown(socket.SHUT_RDWR)
                        request.close()
                        self._logger.info('closed socket from {}', client_addr)
                    except OSError as ex:
                        # the connection is already closed
                        continue
                    except Exception as ex:
                        self._logger.warn ('error closing socket from {}: {}', client_addr, ex)
            time.sleep(0.1)
            for thr in self.httpd.threads:
                thread, request, client_addr = thr
                if thread.is_alive():
                    thread.join(1)
                if thread.is_alive():
                    self._logger.warn('could not end client thread connected from {}', client_addr)
            del self.httpd.threads[:]



import copy
import http.client
import socket
import time
import traceback
import typing
import urllib
import uuid
from collections import defaultdict
from collections import deque

from lxml import etree as etree_

import sdc11073.certloader
from .exceptions import DeliveryModeRequestedUnavailableError
from .exceptions import InvalidMessageError
from .. import isoduration
from .. import loghelper
from .. import multikey
from .. import observableproperties
from .. import pysoap
from .. import xmlparsing
from ..compression import CompressionHandler
from ..namespaces import DocNamespaceHelper
from ..namespaces import Prefix_Namespace as Prefix
from ..namespaces import msgTag
from ..namespaces import nsmap
from ..namespaces import wsaTag
from ..namespaces import wseTag
from ..namespaces import xmlTag

WsAddress = pysoap.soapenvelope.WsAddress
Soap12Envelope = pysoap.soapenvelope.Soap12Envelope

MAX_ROUNDTRIP_VALUES = 20


class _RoundTripData(object):
    def __init__(self, values, abs_max):
        if values:
            self.values = list(values)  # make a copy
            self.min = min(values)
            self.max = max(values)
            self.avg = sum(values) / len(values)
            self.abs_max = abs_max
        else:
            self.values = None
            self.min = None
            self.max = None
            self.avg = None
            self.abs_max = None

    def __repr__(self):
        return 'min={:.4f} max={:.4f} avg={:.4f} absmax={:.4f}'.format(self.min, self.max, self.avg, self.abs_max)


class _DevSubscription(object):
    MAX_NOTIFY_ERRORS = 1
    IDENT_TAG = etree_.QName('http.local.com', 'MyDevIdentifier')

    def __init__(self, mode, base_urls, notifyToAddress, notifyRefNode, endToAddress, endToRefNode, expires,
                 max_subscription_duration, filter_, sslContext,
                 acceptedEncodings):  # pylint:disable=too-many-arguments
        """
        :param notifyToAddress: dom node of Subscribe Request
        :param endToAddress: dom node of Subscribe Request
        :param expires: seconds as float
        :param filter: a space separated list of actions, or only one action
        """
        self.mode = mode
        self.base_urls = base_urls
        self.notifyToAddress = notifyToAddress
        self._url = urllib.parse.urlparse(notifyToAddress)

        self.notifyRefNodes = []
        if notifyRefNode is not None:
            self.notifyRefNodes = list(notifyRefNode)  # all children

        self.endToAddress = endToAddress
        self.endToRefNodes = []
        if endToRefNode is not None:
            self.endToRefNodes = list(endToRefNode)  # all children
        self.my_identifier = etree_.Element(self.IDENT_TAG)
        self.my_identifier.text = uuid.uuid4().urn

        self._max_subscription_duration = max_subscription_duration
        self._started = None
        self._expireseconds = None
        self.renew(expires)  # sets self._started and self._expireseconds
        self._filters = filter_.split()
        self._sslContext = sslContext

        self._acceptedEncodings = acceptedEncodings  # these encodings does the other side accept
        self._soapClient = None

        self._notifyErrors = 0
        self._is_closed = False
        self._isConnectionError = False
        self.last_roundtrip_times = deque(
            maxlen=MAX_ROUNDTRIP_VALUES)  # a list of last n roundtrip times for notifications
        self.max_roundtrip_time = 0

    def setSoapClient(self, soapClient):
        self._soapClient = soapClient

    def renew(self, expires):
        self._started = time.monotonic()
        if expires:
            self._expireseconds = min(expires, self._max_subscription_duration)
        else:
            self._expireseconds = self._max_subscription_duration

    @property
    def netloc(self):
        return self._url.netloc

    @property
    def soapClient(self):
        return self._soapClient

    @property
    def remainingSeconds(self):
        duration = int(self._expireseconds - (time.monotonic() - self._started))
        return 0 if duration < 0 else duration

    @property
    def expireString(self):
        return isoduration.durationString(self.remainingSeconds)

    @property
    def hasDeliveryFailure(self):
        return self._notifyErrors >= self.MAX_NOTIFY_ERRORS

    @property
    def hasConnectionError(self):
        return self._isConnectionError

    @property
    def isValid(self):
        if self._is_closed:
            return False
        if self._soapClient is None or self._soapClient.isClosed():
            return False
        return self.remainingSeconds > 0 and not self.hasDeliveryFailure

    def matches(self, action):
        action = action.strip()  # just to be sure there are no spaces....
        for f in self._filters:
            if f.endswith(action):
                return True
        return False

    def _mkNotificationReport(self, soapEnvelope, action):
        addr = pysoap.soapenvelope.WsAddress(to=self.notifyToAddress,
                                             action=action,
                                             from_=None,
                                             replyTo=None,
                                             faultTo=None,
                                             referenceParametersNode=None)
        soapEnvelope.setAddress(addr)
        for identNode in self.notifyRefNodes:
            identNode_ = copy.copy(identNode)
            # mandatory attribute acc. to ws_addressing SOAP Binding (https://www.w3.org/TR/2006/REC-ws-addr-soap-20060509/)
            identNode_.set(wsaTag('IsReferenceParameter'), 'true')
            soapEnvelope.addHeaderElement(identNode_)
        return soapEnvelope

    def _mkEndReport(self, soapEnvelope, action):
        to_addr = self.endToAddress or self.notifyToAddress
        addr = pysoap.soapenvelope.WsAddress(to=to_addr,
                                             action=action,
                                             from_=None,
                                             replyTo=None,
                                             faultTo=None,
                                             referenceParametersNode=None)
        soapEnvelope.setAddress(addr)
        ref_nodes = self.endToRefNodes or self.notifyRefNodes
        for identNode in ref_nodes:
            identNode_ = copy.copy(identNode)
            # mandatory attribute acc. to ws_addressing SOAP Binding (https://www.w3.org/TR/2006/REC-ws-addr-soap-20060509/)
            identNode_.set(wsaTag('IsReferenceParameter'), 'true')
            soapEnvelope.addHeaderElement(identNode_)
        return soapEnvelope

    def sendNotificationReport(self, bodyNode, action, doc_nsmap):
        if not self.isValid:
            return
        soapEnvelope = pysoap.soapenvelope.Soap12Envelope(doc_nsmap)
        soapEnvelope.addBodyElement(bodyNode)
        rep = self._mkNotificationReport(soapEnvelope, action)
        try:
            roundtrip_timer = observableproperties.SingleValueCollector(self._soapClient, 'roundtrip_time')

            self._soapClient.postSoapEnvelopeTo(self._url.path, rep, responseFactory=None,
                                                msg='sendNotificationReport {}'.format(action))
            try:
                roundtrip_time = roundtrip_timer.result(0)
                self.last_roundtrip_times.append(roundtrip_time)
                self.max_roundtrip_time = max(self.max_roundtrip_time, roundtrip_time)
            except observableproperties.TimeoutError:
                pass
            self._notifyErrors = 0
            self._isConnectionError = False
        except pysoap.soapclient.HTTPReturnCodeError:
            self._notifyErrors += 1
            raise
        except Exception:  # any other exception is handled as an unreachable location (disconnected)
            self._notifyErrors += 1
            self._isConnectionError = True
            raise

    def sendNotificationEndMessage(self, action, code='SourceShuttingDown', reason='Event source going off line.'):
        if not self.isValid:
            return
        doc_nsmap = DocNamespaceHelper().docNssmap
        my_addr = '{}:{}/{}'.format(self.base_urls[0].scheme, self.base_urls[0].netloc, self.base_urls[0].path)

        soapEnvelope = pysoap.soapenvelope.Soap12Envelope(doc_nsmap)

        subscriptionEndNode = etree_.Element(wseTag('SubscriptionEnd'),
                                             nsmap=Prefix.partialMap(Prefix.WSE, Prefix.WSA, Prefix.XML))
        subscriptionManagerNode = etree_.SubElement(subscriptionEndNode, wseTag('SubscriptionManager'))
        # child of Subscriptionmanager is the endpoint reference of the subscription manager (wsa:EndpointReferenceType)
        referenceParametersNode = etree_.Element(wsaTag('ReferenceParameters'))
        referenceParametersNode.append(copy.copy(self.my_identifier))
        epr = pysoap.soapenvelope.WsaEndpointReferenceType(address=my_addr,
                                                           referenceParametersNode=referenceParametersNode)
        epr.asEtreeSubNode(subscriptionManagerNode)

        # remark: optionally one could add own address and identifier here ...
        statusNode = etree_.SubElement(subscriptionEndNode, wseTag('Status'))
        statusNode.text = 'wse:{}'.format(code)
        reasonNode = etree_.SubElement(subscriptionEndNode, wseTag('Reason'),
                                       attrib={xmlTag('lang'): 'en-US'})
        reasonNode.text = reason

        soapEnvelope.addBodyElement(subscriptionEndNode)
        rep = self._mkEndReport(soapEnvelope, action)
        try:
            self._soapClient.postSoapEnvelopeTo(self._url.path, rep, responseFactory=None,
                                                msg='sendNotificationEndMessage {}'.format(action))
            self._notifyErrors = 0
            self._isConnectionError = False
            self._is_closed = True
        except Exception:
            # it does not matter that we could not send the message - end is end ;)
            pass

    def close(self):
        self._is_closed = True

    def isClosed(self):
        return self._is_closed

    def __repr__(self):
        try:
            refIdent = ', '.join([node.text for node in self.notifyRefNodes])
        except TypeError:
            refIdent = '<unknown>'
        return 'Subscription(notifyTo={} idnt={}, my_identifier={}, expires={}, filter={})'.format(self.notifyToAddress,
                                                                                                   refIdent,
                                                                                                   self.my_identifier.text,
                                                                                                   self.remainingSeconds,
                                                                                                   xmlparsing.shortFilterString(
                                                                                                       self._filters))

    @classmethod
    def fromSoapEnvelope(cls, soapEnvelope, sslContext, acceptedEncodings, max_subscription_duration, base_urls):
        endToAddress = None
        endToRefNode = []
        endToAddresses = soapEnvelope.bodyNode.xpath('wse:Subscribe/wse:EndTo', namespaces=nsmap)
        if len(endToAddresses) == 1:
            endToNode = endToAddresses[0]
            endToAddress = endToNode.xpath('wsa:Address/text()', namespaces=nsmap)
            if not endToAddress:
                raise InvalidMessageError(request=soapEnvelope, detail="wse:Subscribe/wse:EndTo/wsa:Address not set")
            endToAddress = endToAddress[0]
            endToRefNode = endToNode.find('wsa:ReferenceParameters', namespaces=nsmap)

        # determine (mandatory) notification address
        deliveryNode = soapEnvelope.bodyNode.xpath('wse:Subscribe/wse:Delivery', namespaces=nsmap)[0]
        notifyToNode = deliveryNode.find('wse:NotifyTo', namespaces=nsmap)
        if notifyToNode is None:
            raise InvalidMessageError(request=soapEnvelope, detail="wse:Subscribe/wse:Delivery/wse:NotifyTo not set")
        notifyToAddress = notifyToNode.xpath('wsa:Address/text()', namespaces=nsmap)[0]
        notifyRefNode = notifyToNode.find('wsa:ReferenceParameters', namespaces=nsmap)

        mode = deliveryNode.get('Mode', pysoap.soapenvelope.MODE_PUSH)
        if mode != pysoap.soapenvelope.MODE_PUSH:
            raise DeliveryModeRequestedUnavailableError(request=soapEnvelope)

        expiresNodes = soapEnvelope.bodyNode.xpath('wse:Subscribe/wse:Expires/text()', namespaces=nsmap)
        expires = isoduration.parse_duration(str(expiresNodes[0])) if expiresNodes else None

        filter_ = soapEnvelope.bodyNode.xpath('wse:Subscribe/wse:Filter/text()', namespaces=nsmap)
        if not filter_:
            raise InvalidMessageError(request=soapEnvelope, detail="wse:Subscribe/wse:Filter not set")
        filter_ = filter_[0]

        return cls(str(mode), base_urls, notifyToAddress, notifyRefNode, endToAddress, endToRefNode,
                   expires, max_subscription_duration, str(filter_), sslContext, acceptedEncodings)

    def get_roundtrip_stats(self):
        if len(self.last_roundtrip_times) > 0:
            return _RoundTripData(self.last_roundtrip_times, self.max_roundtrip_time)
        else:
            return _RoundTripData(None, None)

    def short_filter_names(self):
        return tuple([f.split('/')[-1] for f in self._filters])


class SubscriptionsManager(object):
    BodyNodePrefixes = [Prefix.PM, Prefix.MSG, Prefix.XSI, Prefix.EXT, Prefix.XML]
    NotificationPrefixes = [Prefix.S12, Prefix.WSA, Prefix.WSE]
    DEFAULT_MAX_SUBSCR_DURATION = 7200  # max. possible duration of a subscription

    _ssl_context_container: typing.Optional[sdc11073.certloader.SSLContextContainer]

    def __init__(self, ssl_context_container, sdc_definitions, supportedEncodings,
                 max_subscription_duration=None, log_prefix=None, chunked_messages=False):
        self._ssl_context_container = ssl_context_container
        self.sdc_definitions = sdc_definitions
        self.log_prefix = log_prefix
        self._logger = loghelper.getLoggerAdapter('sdc.device.subscrMgr', self.log_prefix)
        self._chunked_messages = chunked_messages
        self.soapClients = {}  # key: net location, value soapClient instance
        self._supportedEncodings = supportedEncodings
        self._max_subscription_duration = max_subscription_duration or self.DEFAULT_MAX_SUBSCR_DURATION
        self._subscriptions = multikey.MultiKeyLookup()
        self._subscriptions.addIndex('identifier', multikey.UIndexDefinition(lambda obj: obj.my_identifier.text))
        self._subscriptions.addIndex('netloc', multikey.IndexDefinition(
            lambda obj: obj._url.netloc))  # pylint:disable=protected-access
        self.base_urls = None

    def setBaseUrls(self, base_urls):
        self.base_urls = base_urls

    def onSubscribeRequest(self, httpHeader, soapEnvelope, epr_path):
        acceptedEncodings = CompressionHandler.parseHeader(httpHeader.get('Accept-Encoding'))
        s = _DevSubscription.fromSoapEnvelope(
            soapEnvelope,
            self._ssl_context_container.client_context if self._ssl_context_container else None,
            acceptedEncodings,
            self._max_subscription_duration,
            self.base_urls)
        # assign a soap client
        key = s._url.netloc  # pylint:disable=protected-access
        soapClient = self.soapClients.get(key)
        if soapClient is None:
            soapClient = pysoap.soapclient.SoapClient(
                key,
                loghelper.getLoggerAdapter('sdc.device.soap', self.log_prefix),
                sslContext=self._ssl_context_container.client_context if self._ssl_context_container else None,
                sdc_definitions=self.sdc_definitions,
                supportedEncodings=self._supportedEncodings,
                requestEncodings=acceptedEncodings,
                chunked_requests=self._chunked_messages)
            self.soapClients[key] = soapClient
        s.setSoapClient(soapClient)
        with self._subscriptions.lock:
            self._subscriptions.addObject(s)
        self._logger.info('new {}', s)

        response = Soap12Envelope(Prefix.partialMap(*self.NotificationPrefixes))
        replyAddress = soapEnvelope.address.mkReplyAddress(
            'http://schemas.xmlsoap.org/ws/2004/08/eventing/SubscribeResponse')
        response.addHeaderObject(replyAddress)
        subscribeResponseNode = etree_.Element(wseTag('SubscribeResponse'))
        subscriptionManagerNode = etree_.SubElement(subscribeResponseNode, wseTag('SubscriptionManager'))
        # child of Subscriptionmanager is the endpoint reference of the subscription manager (wsa:EndpointReferenceType)
        referenceParametersNode = etree_.Element(wsaTag('ReferenceParameters'))
        referenceParametersNode.append(copy.copy(s.my_identifier))
        if epr_path.startswith('/'):
            epr_path = epr_path[1:]
        my_addr = '{}://{}/{}'.format(self.base_urls[0].scheme, self.base_urls[0].netloc, epr_path)
        epr = pysoap.soapenvelope.WsaEndpointReferenceType(address=my_addr,
                                                           referenceParametersNode=referenceParametersNode)
        epr.asEtreeSubNode(subscriptionManagerNode)
        expiresNode = etree_.SubElement(subscribeResponseNode, wseTag('Expires'))
        expiresNode.text = s.expireString  # simply confirm request
        response.addBodyElement(subscribeResponseNode)
        self._logger.debug('onSubscribeRequest returns {}', lambda: response.as_xml(pretty=False))
        return response

    def onUnsubscribeRequest(self, soapEnvelope):
        ident = soapEnvelope.headerNode.find(_DevSubscription.IDENT_TAG, namespaces=nsmap)
        if ident is not None:
            identtext = ident.text
            s = self._subscriptions.identifier.getOne(identtext, allowNone=True)
            if s is None:
                self._logger.warn('unsubscribe: no object found for id={}', identtext)
            else:
                s.close()
                with self._subscriptions.lock:
                    self._subscriptions.removeObject(s)
                self._logger.info('unsubscribe: object found and removed (Xaddr = {}, filter = {})', s.notifyToAddress,
                                  s._filters)  # pylint: disable=protected-access
                # now check if we can close the soap client
                key = s._url.netloc  # pylint: disable=protected-access
                subscriptionsWithSameSoapClient = self._subscriptions.netloc.get(key, [])
                if len(subscriptionsWithSameSoapClient) == 0:
                    self.soapClients[key].close()
                    del self.soapClients[key]
                    self._logger.info('unsubscribe: closed soap client to {})', key)
        else:
            self._logger.error('unsubscribe request did not contain an identifier!!!: {}',
                               soapEnvelope.as_xml(pretty=True))

        response = Soap12Envelope(nsmap)
        replyAddress = soapEnvelope.address.mkReplyAddress(
            'http://schemas.xmlsoap.org/ws/2004/08/eventing/UnsubscribeResponse')
        response.addHeaderObject(replyAddress)
        # response has empty body
        return response

    def notifyOperation(self, mdib_version_group, transactionId, operationHandleRef, operationState, error=None,
                        errorMessage=None):
        self._logger.info(
            'notifyOperation transaction={} operationHandleRef={}, operationState={}, error={}, errorMessage={}',
            transactionId, operationHandleRef, operationState, error, errorMessage)
        action = self.sdc_definitions.Actions.OperationInvokedReport
        subscribers = self._getSubscriptionsForAction(action)

        bodyNode = etree_.Element(msgTag('OperationInvokedReport'),
                                  nsmap=Prefix.partialMap(Prefix.MSG, Prefix.PM))
        mdib_version_group.update_node(bodyNode)
        reportPartNode = etree_.SubElement(bodyNode,
                                           msgTag('ReportPart'),
                                           attrib={'OperationHandleRef': operationHandleRef})
        invocationInfoNode = etree_.SubElement(reportPartNode, msgTag('InvocationInfo'))
        invocationSourceNode = etree_.SubElement(reportPartNode, msgTag('InvocationSource'),
                                                 attrib={'Root': Prefix.SDC.namespace,
                                                         'Extension': 'AnonymousSdcParticipant'})
        # implemented only SDC R0077 for value of invocationSourceNode:
        # Root =  "http://standards.ieee.org/downloads/11073/11073-20701-2018"
        # Extension = "AnonymousSdcParticipant".
        # a known participant (R0078) is currently not supported
        # ToDo: implement R0078
        transactionIdNode = etree_.SubElement(invocationInfoNode, msgTag('TransactionId'))
        transactionIdNode.text = str(transactionId)
        operationStateNode = etree_.SubElement(invocationInfoNode, msgTag('InvocationState'))
        operationStateNode.text = str(operationState)
        if error is not None:
            errorNode = etree_.SubElement(invocationInfoNode, msgTag('InvocationError'))
            errorNode.text = str(error)
        if errorMessage is not None:
            errorMessageNode = etree_.SubElement(invocationInfoNode, msgTag('InvocationErrorMessage'))
            errorMessageNode.text = str(errorMessage)

        for s in subscribers:
            self._logger.info('notifyOperation: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, Prefix.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def onGetStatusRequest(self, soapEnvelope):
        self._logger.debug('onGetStatusRequest {}', lambda: soapEnvelope.as_xml(pretty=True))
        subscr = self._getSubscriptionforRequest(soapEnvelope)
        if subscr is None:
            response = pysoap.soapenvelope.SoapFault(soapEnvelope,
                                                     code='Receiver',
                                                     reason='unknown Subscription identifier',
                                                     subCode=wseTag('InvalidMessage')
                                                     )

        else:
            response = Soap12Envelope(Prefix.partialMap(*self.NotificationPrefixes))
            replyAddress = soapEnvelope.address.mkReplyAddress(
                'http://schemas.xmlsoap.org/ws/2004/08/eventing/GetStatusResponse')
            response.addHeaderObject(replyAddress)
            renewResponseNode = etree_.Element(wseTag('GetStatusResponse'))
            expiresNode = etree_.SubElement(renewResponseNode, wseTag('Expires'))
            expiresNode.text = subscr.expireString  # simply confirm request
            response.addBodyElement(renewResponseNode)
        return response

    def onRenewRequest(self, soapEnvelope):
        identifierNode = soapEnvelope.headerNode.find(_DevSubscription.IDENT_TAG, namespaces=nsmap)
        expires = soapEnvelope.bodyNode.xpath('wse:Renew/wse:Expires/text()', namespaces=nsmap)
        if len(expires) == 0:
            expires = None
            self._logger.debug('onRenewRequest: no requested duration found, allowing max. ',
                               lambda: soapEnvelope.as_xml(pretty=True))
        else:
            expires = isoduration.parse_duration(str(expires[0]))
            self._logger.debug('onRenewRequest {} seconds', expires)

        subscr = self._getSubscriptionforRequest(soapEnvelope)
        if subscr is None:
            response = pysoap.soapenvelope.SoapFault(soapEnvelope,
                                                     code='Receiver',
                                                     reason='unknown Subscription identifier',
                                                     subCode=wseTag('UnableToRenew')
                                                     )

        else:
            subscr.renew(expires)

            response = Soap12Envelope(Prefix.partialMap(*self.NotificationPrefixes))
            replyAddress = soapEnvelope.address.mkReplyAddress(
                'http://schemas.xmlsoap.org/ws/2004/08/eventing/RenewResponse')
            response.addHeaderObject(replyAddress)
            renewResponseNode = etree_.Element(wseTag('RenewResponse'))
            expiresNode = etree_.SubElement(renewResponseNode, wseTag('Expires'))
            expiresNode.text = subscr.expireString
            response.addBodyElement(renewResponseNode)
        return response

    def sendEpisodicMetricReport(self, updatedMetricStates, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.EpisodicMetricReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending episodic metric report {}', updatedMetricStates)
        bodyNode = etree_.Element(msgTag('EpisodicMetricReport'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)
        reportPartNode = etree_.SubElement(bodyNode, msgTag('ReportPart'))

        for s in updatedMetricStates:
            stateNode = s.mkStateNode(msgTag('MetricState'))
            reportPartNode.append(stateNode)

        for s in subscribers:
            self._logger.debug('sendEpisodicMetricReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def sendPeriodicMetricReport(self, updatedMetricStatesList, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.PeriodicMetricReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending periodic metric report, contains last {} episodic updates',
                           len(updatedMetricStatesList))
        bodyNode = etree_.Element(msgTag('PeriodicMetricReport'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)
        bodyNode.set('MdibVersion', str(updatedMetricStatesList[-1].mdib_version))  # use mdib version from latest part
        for part in updatedMetricStatesList:
            reportPartNode = etree_.SubElement(bodyNode, msgTag('ReportPart'))
            for s in part.states:
                stateNode = s.mkStateNode(msgTag('MetricState'))
                reportPartNode.append(stateNode)

        for s in subscribers:
            self._logger.debug('sendPeriodicMetricReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def sendEpisodicOperationalStateReport(self, updatedStates, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.EpisodicOperationalStateReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending episodic operational state report {}', updatedStates)
        bodyNode = etree_.Element(msgTag('EpisodicOperationalStateReport'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)

        reportPartNode = etree_.SubElement(bodyNode, msgTag('ReportPart'))

        for s in updatedStates:
            stateNode = s.mkStateNode(msgTag('OperationState'))
            reportPartNode.append(stateNode)

        for s in subscribers:
            self._logger.debug('sendEpisodicOperationalStateReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def sendPeriodicOperationalStateReport(self, updatedStatesList, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.PeriodicOperationalStateReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending periodic operational state report, contains last {} episodic updates',
                           len(updatedStatesList))
        bodyNode = etree_.Element(msgTag('PeriodicOperationalStateReport'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)
        bodyNode.set('MdibVersion', str(updatedStatesList[-1].mdib_version))  # use mdib version from latest part
        for part in updatedStatesList:
            reportPartNode = etree_.SubElement(bodyNode, msgTag('ReportPart'))
            for s in part.states:
                stateNode = s.mkStateNode(msgTag('OperationState'))
                reportPartNode.append(stateNode)

        for s in subscribers:
            self._logger.debug('sendPeriodicOperationalStateReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def sendEpisodicAlertReport(self, updatedAlertStates, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.EpisodicAlertReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending episodic alert report {}', updatedAlertStates)
        bodyNode = etree_.Element(msgTag('EpisodicAlertReport'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)
        reportPartNode = etree_.SubElement(bodyNode, msgTag('ReportPart'))

        for s in updatedAlertStates:
            stateNode = s.mkStateNode(msgTag('AlertState'))
            reportPartNode.append(stateNode)

        for s in subscribers:
            self._logger.debug('sendEpisodicAlertReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def sendPeriodicAlertReport(self, updatedStatesList, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.PeriodicAlertReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending periodic alert report, contains last {} episodic updates', len(updatedStatesList))
        bodyNode = etree_.Element(msgTag('PeriodicAlertReport'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)
        bodyNode.set('MdibVersion', str(updatedStatesList[-1].mdib_version))  # use mdib version from latest part
        for part in updatedStatesList:
            reportPartNode = etree_.SubElement(bodyNode, msgTag('ReportPart'))
            for s in part.states:
                stateNode = s.mkStateNode(msgTag('AlertState'))
                reportPartNode.append(stateNode)

        for s in subscribers:
            self._logger.debug('sendPeriodicAlertReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def sendEpisodicComponentStateReport(self, updatedComponentStates, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.EpisodicComponentReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending episodic component report {}', updatedComponentStates)
        bodyNode = etree_.Element(msgTag('EpisodicComponentReport'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)
        reportPartNode = etree_.SubElement(bodyNode, msgTag('ReportPart'))

        for s in updatedComponentStates:
            stateNode = s.mkStateNode(msgTag('ComponentState'))
            reportPartNode.append(stateNode)

        for s in subscribers:
            self._logger.debug('sendEpisodicComponentStateReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def sendPeriodicComponentStateReport(self, updatedStatesList, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.PeriodicComponentReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending periodic component report, contains last {} episodic updates',
                           len(updatedStatesList))
        bodyNode = etree_.Element(msgTag('PeriodicComponentReport'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)
        bodyNode.set('MdibVersion', str(updatedStatesList[-1].mdib_version))  # use mdib version from latest part
        for part in updatedStatesList:
            reportPartNode = etree_.SubElement(bodyNode, msgTag('ReportPart'))
            for s in part.states:
                stateNode = s.mkStateNode(msgTag('ComponentState'))
                reportPartNode.append(stateNode)

        for s in subscribers:
            self._logger.debug('sendPeriodicComponentStateReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def sendEpisodicContextReport(self, updatedContextStates, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.EpisodicContextReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending episodic context report {}', updatedContextStates)
        bodyNode = etree_.Element(msgTag('EpisodicContextReport'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)
        reportPartNode = etree_.SubElement(bodyNode, msgTag('ReportPart'))

        for s in updatedContextStates:
            stateNode = s.mkStateNode(msgTag('ContextState'))
            reportPartNode.append(stateNode)

        for s in subscribers:
            self._logger.info('sendEpisodicContextReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def sendPeriodicContextReport(self, updatedStatesList, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.PeriodicContextReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending periodic context report, contains last {} episodic updates', len(updatedStatesList))
        bodyNode = etree_.Element(msgTag('PeriodicContextReport'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)
        bodyNode.set('MdibVersion', str(updatedStatesList[-1].mdib_version))  # use mdib version from latest part
        for part in updatedStatesList:
            reportPartNode = etree_.SubElement(bodyNode, msgTag('ReportPart'))
            for s in part.states:
                stateNode = s.mkStateNode(msgTag('ContextState'))
                reportPartNode.append(stateNode)

        for s in subscribers:
            self._logger.debug('sendPeriodicContextStateReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def sendRealtimeSamplesReport(self, updatedRealTimeSampleStates, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.Waveform
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending real time samples report {}', updatedRealTimeSampleStates)
        bodyNode = etree_.Element(msgTag('WaveformStream'),
                                  nsmap=nsmapper.partialMap(*self.BodyNodePrefixes))
        mdib_version_group.update_node(bodyNode)

        for s in updatedRealTimeSampleStates:
            stateNode = s.mkStateNode(msgTag('State'))
            bodyNode.append(stateNode)
        for s in subscribers:
            self._logger.debug('sendRealtimeSamplesReport: sending report to {}', s.notifyToAddress)
            self._sendNotificationReport(s, bodyNode, action, nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def endAllSubscriptions(self, sendSubscriptionEnd):
        action = self.sdc_definitions.Actions.SubscriptionEnd
        with self._subscriptions.lock:
            if sendSubscriptionEnd:
                for s in self._subscriptions.objects:
                    s.sendNotificationEndMessage(action)
            self._subscriptions.clear()

    def _mkDescriptorUpdatesReportPart(self, parentNode, modificationtype, descriptors, updated_states):
        """ Helper that creates ReportPart."""
        # This method creates one ReportPart for every descriptor.
        # An optimization is possible by grouping all descriptors with the same parent handle into one ReportPart.
        # This is not implemented, and I think it is not needed.
        for descrContainer in descriptors:
            reportPart = etree_.SubElement(parentNode, msgTag('ReportPart'),
                                           attrib={'ModificationType': modificationtype})
            if descrContainer.parentHandle is not None:  # only Mds can have None
                reportPart.set('ParentDescriptor', descrContainer.parentHandle)
            node = descrContainer.mkDescriptorNode(tag=msgTag('Descriptor'))
            reportPart.append(node)
            relatedStateContainers = [s for s in updated_states if s.descriptorHandle == descrContainer.handle]
            for stateContainer in relatedStateContainers:
                node = stateContainer.mkStateNode(msgTag('State'))
                reportPart.append(node)

    def sendDescriptorUpdates(self, updated, created, deleted, updated_states, nsmapper, mdib_version_group):
        action = self.sdc_definitions.Actions.DescriptionModificationReport
        subscribers = self._getSubscriptionsForAction(action)
        if not subscribers:
            return
        self._logger.debug('sending DescriptionModificationReport upd={} crt={} del={}', updated, created, deleted)
        bodyNode = etree_.Element(msgTag('DescriptionModificationReport'),
                                  nsmap=nsmapper.partialMap(Prefix.MSG, Prefix.PM))
        mdib_version_group.update_node(bodyNode)
        self._mkDescriptorUpdatesReportPart(bodyNode, 'Upt', updated, updated_states)
        self._mkDescriptorUpdatesReportPart(bodyNode, 'Crt', created, updated_states)
        self._mkDescriptorUpdatesReportPart(bodyNode, 'Del', deleted, updated_states)

        for s in subscribers:
            self._sendNotificationReport(s, bodyNode, action,
                                         nsmapper.partialMap(*self.NotificationPrefixes))
        self._doHousekeeping()

    def _sendNotificationReport(self, subscription, bodyNode, action, doc_nsmap):
        try:
            subscription.sendNotificationReport(bodyNode, action, doc_nsmap)
        except pysoap.soapclient.HTTPReturnCodeError as ex:
            # this is an error related to the connection => log error and continue
            self._logger.error('could not send notification report: HTTP status= {}, reason={}, {}', ex.status,
                               ex.reason, subscription)
        except http.client.NotConnected as ex:
            # this is an error related to the connection => log error and continue
            self._logger.error('could not send notification report: {!r}:  subscr = {}', ex, subscription)
        except socket.timeout as ex:
            # this is an error related to the connection => log error and continue
            self._logger.error('could not send notification report error= {!r}: {}', ex, subscription)
        except etree_.DocumentInvalid as ex:
            # this is an error related to the document, it cannot be sent to any subscriber => re-raise
            self._logger.error('Invalid Document: {!r}\n{}', ex, etree_.tostring(bodyNode))
            raise
        except Exception as ex:
            # this should never happen! => re-raise
            self._logger.error('could not send notification report: {}: \n{}', traceback.format_exc(), subscription)

    def _getSubscriptionsForAction(self, action):
        with self._subscriptions.lock:
            return [s for s in self._subscriptions.objects if s.matches(action)]

    def _getSubscriptionforRequest(self, soapEnvelope):
        request_name = soapEnvelope.bodyNode[0].tag
        identifierNode = soapEnvelope.headerNode.find(_DevSubscription.IDENT_TAG, namespaces=nsmap)
        if identifierNode is None:
            raise RuntimeError('no Identifier found in {} ', request_name)
        else:
            identifier = identifierNode.text
        with self._subscriptions.lock:
            subscr = [s for s in self._subscriptions.objects if s.my_identifier.text == identifier]
        if len(subscr) > 1:
            raise RuntimeError('Have {} subscriptions with identifier "{}"!'.format(len(subscr), identifier))
        elif len(subscr) == 0:
            self._logger.error('on {}: unknown Subscription identifier "{}"', request_name, identifier)
            return
        return subscr[0]

    def _doHousekeeping(self):
        """ remove expired or invalid subscriptions"""
        with self._subscriptions._lock:  # pylint: disable=protected-access
            crap = [s for s in self._subscriptions.objects if not s.isValid]
        unreachable_netlocs = []
        for c in crap:
            if c.hasConnectionError:
                # the network location is unreachable, we can remove all subscriptions that use this location
                unreachable_netlocs.append(c.soapClient.netloc)
                try:
                    c.soapClient.close()
                except:
                    self._logger.error('error in soapClient.close(): {}', traceback.format_exc())

            self._logger.info('deleting {}, errors={}', c, c._notifyErrors)  # pylint: disable=protected-access
            with self._subscriptions.lock:
                self._subscriptions.removeObject(c)

            if c.netloc in self.soapClients:  # remove closed soap client from list
                del self.soapClients[c.netloc]

        # now find all subscriptions that have the same address
        with self._subscriptions._lock:  # pylint: disable=protected-access
            also_unreachable = [s for s in self._subscriptions.objects if
                                s.soapClient is not None and s.netloc in unreachable_netlocs]
            for s in also_unreachable:
                self._logger.info('deleting also subscription {}, same endpoint', s)
                self._subscriptions.removeObject(s)

    def getSubScriptionRoundtripTimes(self):
        """Calculates roundtrip times based on last MAX_ROUNDTRIP_VALUES values.

        @return: a dictionary with key=(<notifyToAddress>, (subscriptionnames)), value = _RoundTripData with members min, max, avg, abs_max, values
        """
        ret = {}
        with self._subscriptions.lock:
            for s in self._subscriptions.objects:
                if s.max_roundtrip_time > 0:
                    ret[(s.notifyToAddress, s.short_filter_names())] = s.get_roundtrip_stats()
        return ret

    def getClientRoundtripTimes(self):
        """Calculates roundtrip times based on last MAX_ROUNDTRIP_VALUES values.

        @return: a dictionary with key=<notifyToAddress>, value = _RoundTripData with members min, max, avg, abs_max, values
        """
        # first step: collect all roundtrip times of subscriptions, group them by notifyToAddress
        tmp = defaultdict(list)
        ret = {}
        with self._subscriptions.lock:
            for s in self._subscriptions.objects:
                if s.max_roundtrip_time > 0:
                    tmp[s.notifyToAddress].append(s.get_roundtrip_stats())
        for k, stats in tmp.items():
            allvalues = []
            for s in stats:
                allvalues.extend(s.values)
            ret[k] = _RoundTripData(allvalues, max([s.max for s in stats]), )
        return ret
